import sys
from time import sleep
from typing import Any

from common import ACCEPTED_ERROR_CODES, APITest
from pydantic import RootModel

TESTED_PATH = "entity/{entity}/_/distinct/{attr}"
TESTED_INT_PATH = TESTED_PATH.format(entity="test_entity_type", attr="test_attr_int")
TESTED_SET_PATH = TESTED_PATH.format(entity="test_entity_type", attr="test_attr_set")

# Pydantic-based type for dictionary
GenericDictModel = RootModel[dict[Any, int]]


class GetDistinctAttrValues(APITest):
    @classmethod
    def setUpClass(cls) -> None:
        dps_attr_value = {
            "test_attr_int": 123,
            "test_attr_set": [123, 456],
        }

        # Push datapoints
        for attr in dps_attr_value:
            res = cls.push_datapoints(
                [
                    {
                        "type": "test_entity_type",
                        "id": "test_entity_id",
                        "attr": attr,
                        "v": dps_attr_value[attr],
                    }
                ]
            )
            print(res.content.decode("utf-8"), file=sys.stderr)
        sleep(9)

        # Make snapshots
        cls.get_request("control/make_snapshots")
        sleep(6)

    def test_unknown_entity_type(self):
        response = self.get_request(TESTED_PATH.format(entity="xyz", attr="test_attr_int"))
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES)

    def test_unknown_attr_name(self):
        response = self.get_request(TESTED_PATH.format(entity="test_entity_type", attr="xyz"))
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES)

    def test_valid_get_int(self):
        self.query_expected_value(
            lambda: self.get_entity_data(TESTED_INT_PATH, GenericDictModel),
            lambda received: "123" in received.root,
        )

    def test_valid_get_set(self):
        self.query_expected_value(
            lambda: self.get_entity_data(TESTED_SET_PATH, GenericDictModel),
            lambda received: "123" in received.root and "456" in received.root,
        )



#!/usr/bin/env python3
"""A utility for running DP3 commands using a CLI."""

import argparse
import sys

from dp3.bin.api import init_parser as init_api_parser
from dp3.bin.api import main as api_main
from dp3.bin.check import init_parser as init_check_parser
from dp3.bin.check import main as check_main
from dp3.bin.config import init_parser as init_config_parser
from dp3.bin.config import main as config_main
from dp3.bin.schema_update import init_parser as init_schema_update_parser
from dp3.bin.schema_update import main as schema_update_main
from dp3.bin.setup import init_parser as init_setup_parser
from dp3.bin.setup import main as setup_main
from dp3.bin.worker import init_parser as init_worker_parser
from dp3.bin.worker import main as worker_main


def init_parser():
    parser = argparse.ArgumentParser(prog="dp3")
    commands = parser.add_subparsers(title="commands", required=True, dest="command")

    worker_parser = commands.add_parser(
        "worker",
        help="Run main worker process of the DP3 platform. ",
        description="Main worker process of the DP3 platform. "
        "There are usually multiple workers running in parallel.",
    )
    init_worker_parser(worker_parser)

    api_parser = commands.add_parser("api", help="Run the DP3 API using uvicorn.")
    init_api_parser(api_parser)

    setup_parser = commands.add_parser("setup", help="Create a DP3 application template.")
    init_setup_parser(setup_parser)

    check_parser = commands.add_parser(
        "check",
        help="Check configuration files for errors.",
        description="Load configuration from given directory and check its validity. "
        "When configuration is OK, program exits immediately with status code 0, "
        "otherwise it prints error messages on stderr and exits with non-zero status.",
    )
    init_check_parser(check_parser)

    config_parser = commands.add_parser(
        "config",
        help="Setup configuration files for a container-less DP3 application deployment.",
        description="Setup configuration files for a container-less DP3 application deployment. ",
    )
    init_config_parser(config_parser)

    schema_update_parser = commands.add_parser(
        "schema-update",
        help="Update the database schema after making conflicting changes to the model.",
        description="Update the database schema after making conflicting changes to the model. ",
    )
    init_schema_update_parser(schema_update_parser)
    return parser


def run():
    parser = init_parser()
    args = parser.parse_args()

    if args.command == "worker":
        worker_main(args)
    elif args.command == "api":
        api_main(args)
    elif args.command == "setup":
        setup_main(args)
    elif args.command == "check":
        check_main(args)
    elif args.command == "config":
        config_main(args)
    elif args.command == "schema-update":
        schema_update_main(args)
    else:
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
"""Simple script to add hashes to master records to allow for easier parallelization.

Intended to be run only once to upgrade a database existing before 08-2023.
"""

import argparse

from pymongo import UpdateOne

from dp3.common.config import ModelSpec, read_config_dir
from dp3.common.task import HASH
from dp3.database.database import EntityDatabase, MongoConfig

# Arguments parser
parser = argparse.ArgumentParser(
    description="Add hashes to master records to allow for easier parallelization in ADiCT."
)
parser.add_argument(
    "--config",
    default="/etc/adict/config",
    help="DP3 config directory (default: /etc/adict/config)",
)
parser.add_argument("-n", default=100, type=int, help="Number of updates to send per request")
args = parser.parse_args()

# Load DP3 configuration
config = read_config_dir(args.config, recursive=True)
model_spec = ModelSpec(config.get("db_entities"))

# Connect to database
connection_conf = MongoConfig.parse_obj(config.get("database", {}))
client = EntityDatabase.connect(connection_conf)
client.admin.command("ping")

db = client[connection_conf.db_name]


def send_updates(entity: str, update_list: list):
    if not update_list:
        return
    res = db[f"{entity}#master"].bulk_write(
        [UpdateOne(_record, {"$set": {"#hash": _hash}}) for _record, _hash in update_list]
    )
    print(res.bulk_api_result)
    update_list.clear()


for entity in model_spec.entities:
    print(entity)
    updates = []
    for record in db[f"{entity}#master"].find({}, projection={"_id": True}):
        updates.append((record, HASH(f"{entity}:{record['_id']}")))
        if len(updates) >= args.n:
            send_updates(entity, updates)
    send_updates(entity, updates)



#!/usr/bin/env python3
"""Converts legacy CSV DataPoint log format to JSON"""
import argparse
import gzip
import ipaddress
import json
import logging
import os
import re
from typing import Any, Callable

import pandas as pd
from dateutil.parser import parse as parsetime
from pydantic import ValidationError

from dp3.common.config import ModelSpec, read_config_dir

re_timestamp = re.compile(
    r"^\d{4}-\d{2}-\d{2}[Tt ]\d{2}:\d{2}:\d{2}(?:\.\d+)?([Zz]|(?:[+-]\d{2}:\d{2}))?$"
)
re_mac = re.compile(r"^([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2})$")
re_array = re.compile(r"^array<(\w+)>$")
re_set = re.compile(r"^set<(\w+)>$")
re_link = re.compile(r"^link<(\w+)>$")
re_dict = re.compile(r"^dict<((\w+\??:\w+,)*(\w+\??:\w+))>$")
re_category = re.compile(
    r"^category<\s*(?P<type>\w+)\s*;\s*(?P<vals>(?:\s*\w+,\s*)*(?:\s*\w+\s*))>$"
)


def valid_ipv4(address):
    try:
        ipaddress.IPv4Address(address)
        return True
    except ValueError:
        return False


def valid_ipv6(address):
    try:
        ipaddress.IPv6Address(address)
        return True
    except ValueError:
        return False


# Validate MAC string
def valid_mac(address):
    return re_mac.match(address)


logging.basicConfig(level=logging.INFO, format="%(name)s [%(levelname)s] %(message)s")

# Dictionary containing conversion functions for primitive data types
CONVERTERS = {
    "tag": lambda v: json.loads(f'{{"v": {v}}}')["v"],
    "binary": lambda v: v.lower() == "true",
    "string": str,
    "int": int,
    "int64": int,
    "float": float,
    "ipv4": lambda v: _pass_valid(valid_ipv4, v),
    "ipv6": lambda v: _pass_valid(valid_ipv6, v),
    "mac": lambda v: _pass_valid(valid_mac, v),
    "time": parsetime,
    "json": json.loads,
}


def get_converter(attr_data_type: str) -> Callable[[str], Any]:
    """Return a function converting a string to given data type."""
    # basic type
    if attr_data_type in CONVERTERS:
        return CONVERTERS[attr_data_type]
    # array<X>, set<X>, dict<X,Y,Z>
    if (
        re.match(re_array, attr_data_type)
        or re.match(re_set, attr_data_type)
        or re.match(re_dict, attr_data_type)
    ):
        return json.loads
    # link<X>
    if re.match(re_link, attr_data_type):
        return str
    # category<X; Y>
    if re.match(re_category, attr_data_type):
        return str
    raise ValueError(f"No conversion function for attribute type '{attr_data_type}'")


def _parse_array_str(string: str, item_type: str) -> list:
    """Parse string containing array of items of given type"""
    conv = CONVERTERS[item_type]
    a = json.loads(string)
    if not isinstance(a, list):
        raise ValueError
    return [conv(item) for item in a]


def _parse_set_str(string: str, item_type: str) -> set:
    conv = CONVERTERS[item_type]
    a = json.loads(string)
    if not isinstance(a, list):
        raise ValueError
    return {conv(item) for item in a}


def _parse_dict_str(string: str, field_types: dict) -> dict:
    o = json.loads(string)
    if not isinstance(o, dict):
        raise ValueError
    return {k: CONVERTERS[field_types[k]](v) for k, v in o.items()}


def _pass_valid(validator_function, value):
    if validator_function(value):
        return value
    raise ValueError(f"The value {value} has invalid format.")


class LegacyDataPointLoader:
    """Loader of datapoint files as written by DP3 API receiver."""

    # Names of columns in datapoint files
    COL_NAMES = ["type", "id", "attr", "t1", "t2", "c", "src", "v"]
    log = logging.getLogger("LegacyDataPointLoader")

    def __init__(self, attr_config_dirname: str):
        """
        Create a datapoint loader.

        attr_config_dirname: Directory with attribute configuration (same as for DP3)
        """
        # Load attribute config
        model_spec = ModelSpec(read_config_dir(attr_config_dirname))

        # Prepare a table for data type conversion
        # (to get data type from model_spec: model_spec[etype]["attribs"][attrname].data_type)
        self.dt_conv = {}  # map (etype,attr_name) -> conversion_function
        for etype, spec in model_spec.items():
            for aname, aspec in spec["attribs"].items():
                data_type = getattr(aspec, "data_type", None)
                converter = json.loads if data_type is None else get_converter(str(data_type))
                self.dt_conv[(etype, aname)] = converter

        self.model_spec = model_spec

    def read_dp_file(self, filename: str) -> pd.DataFrame:
        """
        Read a file with ADiCT/DP3 datapoints into pandas DataFrame.

        Values of attributes in datapoints are validated and converted according
        to the attribute configuration passed to LegacyDataPointLoader constructor.
        """
        open_function = gzip.open if filename.endswith(".gz") else open

        # Reformat datapoints file so "val" containing commas can be read properly.
        #   Replace first 7 commas (i.e. all except those inside "v") with semicolon
        #   Store as temporary file
        tmp_name = (
            f"tmp-{'.'.join(os.path.basename(os.path.normpath(filename)).split(sep='.')[:-1])}"
        )
        with open_function(filename, "rb") as infile, open(tmp_name, "wb") as outfile:
            for line in infile:
                outfile.write(line.replace(b",", b";", 7))
        # Load the converted file
        data = pd.read_csv(
            tmp_name,
            sep=";",
            header=None,
            names=self.COL_NAMES,
            index_col=False,
            converters={"c": float, "v": str},
            escapechar="\\",
            # parse_dates=["t1", "t2"],
            # infer_datetime_format=True,
        )
        # Cleanup
        if os.path.exists(tmp_name):
            os.remove(tmp_name)

        # Convert values to correct types according to model_spec
        def convert_row(row):
            try:
                row[2] = self.dt_conv[(row[0], row[1])](row[2])
            except KeyError as e:
                raise KeyError(f"No converter for {(row[0], row[1])}, with value {row[2]}.") from e
            except ValueError:
                self.log.error("ValueError in conversion, v: %s", row)
                return row
            return row

        attrs = {entity_attr[1] for entity_attr in self.dt_conv}
        conv_vals = data.loc[data["attr"].isin(attrs), ("type", "attr", "v")].apply(
            convert_row, axis=1, raw=True
        )
        if len(conv_vals) != len(data):
            self.log.warning(
                "Dropped %s rows due to missing attributes in config", len(data) - len(conv_vals)
            )
            self.log.info("Missing attrs: %s", [x for x in data["attr"].unique() if x not in attrs])
        data["v"] = conv_vals["v"]
        return data[data["attr"].apply(lambda x: x in attrs)]


def get_valid_path(parser, arg):
    if not os.path.exists(arg):
        parser.error(f"The file {arg} does not exist!")
    else:
        return os.path.abspath(arg)


def get_out_path(in_file_path, output_dir):
    """
    Return output file path based on the input file.
    Parses the date from input filename, fits date into prepared pattern: "dp_log_{date}.json".
    """
    date = in_file_path.split("-")[-1]
    if date.endswith(".gz"):
        date = date[:-3]
    out_filename = f"dp_log_{date}.json"
    return os.path.join(output_dir, out_filename)


def validate_row(row):
    # COL_NAMES = ["type", "id", "attr", "t1", "t2", "c", "src", "v"]

    dp_obj = {
        "etype": row[0],
        "eid": row[1],
        "attr": row[2],
        "t1": row[3],
        "t2": row[4],
        "c": row[5],
        "src": row[6],
        "v": row[7],
    }

    etype = dp_obj["etype"]
    attr = dp_obj["attr"]

    try:
        model_spec.attr(etype, attr).dp_model.model_validate(dp_obj)
    except ValidationError as err:
        print(model_spec.attr(etype, attr))
        print(attr, type(dp_obj["v"]), repr(dp_obj["v"]), dp_obj["t1"], dp_obj["t2"])
        raise err


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Converts legacy CSV DataPoint log format to JSON")
    parser.add_argument(
        "-c",
        "--attr_conf_dir",
        dest="attr_conf_dir",
        required=True,
        help="Path to DP3 entity config",
        type=lambda x: get_valid_path(parser, x),
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        dest="output_dir",
        default=os.path.curdir,
        help="Converted files will be saved to this path",
        type=lambda x: get_valid_path(parser, x),
    )
    parser.add_argument(
        "files", help="Legacy CSV file paths", type=lambda x: get_valid_path(parser, x), nargs="+"
    )
    parser.add_argument(
        "--compress",
        action="store_true",
        default=False,
        help="Compress output file using gzip (.gz)",
    )
    args = parser.parse_args()

    loader = LegacyDataPointLoader(args.attr_conf_dir)
    model_spec = loader.model_spec

    for filename in args.files:
        dp_log = loader.read_dp_file(filename)

        dp_log.apply(validate_row, axis=1, raw=True)

        converted_filename = get_out_path(filename, args.output_dir)
        converted_filename = converted_filename + ".gz" if args.compress else converted_filename
        dp_log.to_json(converted_filename, orient="records", indent=1)



"""
Core module managing links between entities.
"""

import logging
from datetime import datetime, timedelta
from functools import partial

from pymongo import DeleteMany

from dp3.common.attrspec import AttrType
from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import PlatformConfig
from dp3.common.datapoint import DataPointBase, DataPointObservationsBase
from dp3.database.database import EntityDatabase


class LinkManager:
    """Manages the shared Link cache and updates links after entity deletion."""

    def __init__(
        self,
        db: EntityDatabase,
        platform_config: PlatformConfig,
        registrar: CallbackRegistrar,
    ):
        self.log = logging.getLogger("LinkManager")
        self.model_spec = platform_config.model_spec
        self.db = db

        self.cache = self.db.get_module_cache("Link")
        self._setup_cache_indexes()
        self.db.register_on_entity_delete(
            self.remove_link_cache_of_deleted, self.remove_link_cache_of_many_deleted
        )
        self.max_date = datetime.max.replace(tzinfo=None)
        for (entity, attr), spec in self.model_spec.relations.items():
            if spec.t == AttrType.PLAIN:
                if spec.is_iterable:
                    func = self.add_iterable_plain_to_link_cache
                else:
                    func = self.add_plain_to_link_cache
                registrar.register_attr_hook(
                    "on_new_plain", partial(func, spec.relation_to), entity, attr
                )
            elif spec.t == AttrType.OBSERVATIONS:
                if spec.is_iterable:
                    func = self.add_iterable_observation_to_link_cache
                else:
                    func = self.add_observation_to_link_cache
                registrar.register_attr_hook(
                    "on_new_observation",
                    partial(func, spec.relation_to, spec.history_params.post_validity),
                    entity,
                    attr,
                )

    def _setup_cache_indexes(self):
        """Sets up indexes for the cache collection.

        In the collection, these fields are covered by an index:

        * The `to` and `from` fields are used when loading linked entities,
          as well as when removing links of deleted entities
        * The `using_attr` field is used to filter which link attributes are used
        * The `ttl` field serves as a MongoDB expiring collection index
        """
        self.cache.create_index("to", background=True)
        self.cache.create_index("from", background=True)
        self.cache.create_index("using_attr", background=True)
        self.cache.create_index("ttl", expireAfterSeconds=0, background=True)

    def add_plain_to_link_cache(self, etype_to: str, eid: str, dp: DataPointBase):
        self.cache.update_one(
            {
                "to": f"{etype_to}#{dp.v.eid}",
                "from": f"{dp.etype}#{eid}",
                "using_attr": f"{dp.etype}#{dp.attr}",
            },
            {"$max": {"ttl": self.max_date}},
            upsert=True,
        )

    def add_iterable_plain_to_link_cache(self, etype_to: str, eid: str, dp: DataPointBase):
        linked_eids = [v.eid for v in dp.v]
        self.cache.update_many(
            {
                "to": {"$in": [f"{etype_to}#{eid_}" for eid_ in linked_eids]},
                "from": f"{dp.etype}#{eid}",
                "using_attr": f"{dp.etype}#{dp.attr}",
            },
            {"$max": {"ttl": self.max_date}},
            upsert=True,
        )

    def add_observation_to_link_cache(
        self, etype_to: str, post_validity: timedelta, eid: str, dp: DataPointObservationsBase
    ):
        self.cache.update_one(
            {
                "to": f"{etype_to}#{dp.v.eid}",
                "from": f"{dp.etype}#{eid}",
                "using_attr": f"{dp.etype}#{dp.attr}",
            },
            {"$max": {"ttl": dp.t2 + post_validity}},
            upsert=True,
        )

    def add_iterable_observation_to_link_cache(
        self, etype_to: str, post_validity: timedelta, eid: str, dp: DataPointObservationsBase
    ):
        linked_eids = [v.eid for v in dp.v]
        self.cache.update_many(
            {
                "to": {"$in": [f"{etype_to}#{eid_}" for eid_ in linked_eids]},
                "from": f"{dp.etype}#{eid}",
                "using_attr": f"{dp.etype}#{dp.attr}",
            },
            {"$max": {"ttl": dp.t2 + post_validity}},
            upsert=True,
        )

    def remove_link_cache_of_deleted(self, etype: str, eid: str):
        # Delete from master records
        # Get all links to the deleted entity grouped by attribute
        links = self.cache.aggregate(
            [
                {"$match": {"to": f"{etype}#{eid}"}},
                {"$group": {"_id": "$using_attr", "eids_from": {"$push": "$from"}}},
            ]
        )
        # Delete the retrieved references
        for link in links:
            etype_from, attr = link["_id"].split("#", maxsplit=1)
            eids_from = [l_eid.split("#", maxsplit=1)[1] for l_eid in link["eids_from"]]
            self.db.delete_link_dps(etype_from, eids_from, attr, eid)

        # Delete from cache
        self.cache.bulk_write(
            [DeleteMany({"from": f"{etype}#{eid}"}), DeleteMany({"to": f"{etype}#{eid}"})]
        )

    def remove_link_cache_of_many_deleted(self, etype: str, eids: list[str]):
        # Delete from master records
        # Get all links to the deleted entities grouped by attribute
        links = self.cache.aggregate(
            [
                # Match links to the deleted entities
                {"$match": {"to": {"$in": [f"{etype}#{eid}" for eid in eids]}}},
                # Group the links by attribute (which includes the etype_from)
                {
                    "$group": {
                        "_id": "$using_attr",
                        "eids_to": {"$push": "$to"},
                        "eids_from": {"$push": "$from"},
                    }
                },
            ]
        )

        etypes_from = []
        attrs = []
        affected_eids = []
        eids_to = []
        for link in links:
            etype_from, attr = link["_id"].split("#", maxsplit=1)
            etypes_from.append(etype_from)
            attrs.append(attr)
            affected_eids.append([l_eid.split("#", maxsplit=1)[1] for l_eid in link["eids_from"]])
            eids_to.append([l_eid.split("#", maxsplit=1)[1] for l_eid in link["eids_to"]])

        self.db.delete_many_link_dps(etypes_from, affected_eids, attrs, eids_to)

        self.cache.bulk_write(
            [DeleteMany({"from": f"{etype}#{eid}"}) for eid in eids]
            + [DeleteMany({"to": f"{etype}#{eid}"}) for eid in eids]
        )



class SharedFlag:
    """A bool wrapper that can be shared between multiple objects"""

    def __init__(self, flag: bool = False, banner: str = ""):
        self._flag = flag
        self._banner = banner

    def set(self, flag: bool = True):
        """Set the flag to `flag`. True by default."""
        self._flag = flag

    def unset(self):
        """Unset the flag."""
        self._flag = False

    def isset(self):
        """Check if the flag is set."""
        return self._flag

    def __repr__(self):
        if self._banner:
            return f"SharedFlag({repr(self._banner)}, {repr(self._flag)})"
        return f"SharedFlag({repr(self._flag)})"

    def __str__(self):
        return str(self._flag)



#!/usr/bin/env python3
"""Simple datapoint sender script for testing local DP3 instance."""
import gzip
import json
import logging
import os
import time
from argparse import ArgumentParser
from datetime import datetime
from itertools import islice
from queue import Queue
from threading import Event, Thread

import pandas as pd
import requests


def get_valid_path(parser, arg):
    if not os.path.exists(arg):
        parser.error(f"The file {arg} does not exist!")
    else:
        return os.path.abspath(arg)


def get_datapoint_from_row(row):
    dp = dict(row.items())
    if pd.isna(dp["t1"]):
        del dp["t1"]
        del dp["t2"]
        return dp
    dp["t1"] = dp["t1"].strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]

    if pd.isna(dp["t2"]):
        del dp["t2"]
        return dp
    dp["t2"] = dp["t2"].strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
    return dp


def get_shifted_datapoint_from_row(row):
    dp = dict(row.items())
    if pd.isna(dp["t1"]):
        del dp["t1"]
        del dp["t2"]
        return dp
    now = datetime.utcnow()
    shift = now - dp["t1"]
    dp["t1"] = now.strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]

    if pd.isna(shift):
        del dp["t2"]
        return dp
    dp["t2"] = (dp["t2"] + shift).strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
    if not isinstance(dp["v"], dict):
        return dp

    ts_fields = ["time_first", "time_last", "time"]
    for field in ts_fields:
        if field in dp["v"]:
            shifted = []
            for item in dp["v"][field]:
                shifted.append(
                    (datetime.fromisoformat(item) + shift).strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
                )
            dp["v"][field] = shifted
    return dp


def batched(iterable, n):
    """Batch data into tuples of length n. The last batch may be shorter."""
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := list(islice(it, n)):
        yield batch


def cherry_pick_send(datapoints, dp_factory, log, args):
    # Get all datapoints that will be sent.
    attributes = datapoints["attr"].unique()
    log.debug("Found %s attributes", len(datapoints))
    log.debug(attributes)
    dp_list = []
    for attr in attributes:
        selected = datapoints[datapoints["attr"] == attr]
        selected_dps = selected.iloc[: args.count, :].apply(dp_factory, axis=1).to_list()
        log.debug("Selected %s datapoints of attribute %s", len(selected_dps), attr)
        dp_list.extend(selected_dps)

    log.info("Selected %s datapoints", len(dp_list))
    return dp_list


def send_all(datapoints, dp_factory, *_):
    for _i, dp in datapoints.iterrows():
        yield dp_factory(dp)


def worker_thread(i: int, args, stop_event: Event, dp_q: Queue, q_empty: Event, queue: Queue):
    log.debug("Starting")

    while not stop_event.is_set():
        if dp_q.empty():
            q_empty.set()
        dps = dp_q.get()
        if dp_q.empty():
            q_empty.set()

        if dps is None:
            log.debug("Received stop event, exiting.")
            break
        send_dps(i, args, stop_event, dps, queue)


def send_dps(i: int, args, stop_event: Event, datapoints, queue: Queue):
    if args.mode == "cherry-pick":
        dps = cherry_pick_send(datapoints, dp_factory, log, args)
    elif args.mode == "all":
        dps = send_all(datapoints, dp_factory, log, args)
    else:
        raise ValueError(f"Invalid mode selected: {args.mode}")

    request_time = 0
    dps_sent = 0

    datapoints_url = f"{args.endpoint_url}/datapoints"
    for batch in batched(dps, args.chunk):
        if stop_event.is_set():
            log.info("Received stop event, exiting.")
            break
        payload = json.dumps(batch)
        log.debug(payload)
        t_sent = time.time()
        try:
            response = requests.post(url=datapoints_url, json=batch)
            t_received = time.time()
            if response.status_code == requests.codes.ok:
                attributes = {dp["attr"] for dp in batch}
                log.debug(
                    "(%.3fs) %s datapoints of attribute(s) %s OK",
                    t_received - t_sent,
                    len(batch),
                    ", ".join(attributes),
                )
            else:
                log.error("Payload: %s", payload)
                log.error(
                    "Request failed: %s: %s", response.reason, response.content.decode("utf-8")
                )

            dps_sent += len(batch)
            request_time += t_received - t_sent

        except requests.exceptions.ConnectionError as err:
            t_error = time.time()
            attributes = {dp["attr"] for dp in batch}
            log.error(
                "(%.3fs) %s - %s datapoints of attribute(s) %s",
                t_error - t_sent,
                err,
                len(batch),
                ", ".join(attributes),
            )
            stop_event.set()
            time.sleep(0.5)  # Make sure the reader does not clear the empty queue signal at startup
            queue_empty.set()
            break
        except requests.exceptions.InvalidJSONError as err:
            log.exception(err)
            log.error(batch)
        except KeyboardInterrupt:
            log.info("Received user interrupt, exiting.")
            break

    log.info("Sent %s datapoints in %.3fs", dps_sent, request_time)
    queue.put((i, dps_sent, request_time))


def reader_thread(
    file_path: str,
    dp_q: Queue,
    q_empty: Event,
    stop_running: Event,
    buffer_size: int = 1000,
    prefetch_size: int = 4,
):
    startup = True
    try:
        dp_iter = datapoint_reader(file_path, buffer_size)
        while not stop_running.is_set():
            for _i in range(prefetch_size):
                dp_chunk = next(dp_iter)
                log.info("Read %s datapoints (%s)", dp_chunk.shape[0], _i + 1)
                dp_q.put(dp_chunk)
            if startup:
                startup = False
                q_empty.clear()

            q_empty.wait()
            q_empty.clear()

    except KeyboardInterrupt:
        log.info("Received user interrupt, exiting.")
    except Exception as e:
        log.exception(e)

    for _ in range(prefetch_size):
        dp_q.put(None)


def datapoint_reader(file_path: str, buffer_size: int = 10000):
    def yield_transform(buf: list[bytes]):
        return pd.read_json(
            b"".join(buf).decode("utf-8"),
            orient="records",
            convert_dates=["t1", "t2"],
            lines="jsonl" in args.input_file,
        ).fillna(value={"c": 1.0})

    open_fn = open if not file_path.endswith(".gz") else gzip.open

    with open_fn(file_path, "rb") as file:
        buffer = []
        for line in file:
            buffer.append(line)
            if len(buffer) >= buffer_size:
                yield yield_transform(buffer)
                buffer = []
        if buffer:
            yield yield_transform(buffer)


if __name__ == "__main__":
    parser = ArgumentParser(
        description="Simple datapoint sender script for testing local DP3 instance."
    )
    parser.add_argument(
        "input_file",
        help="DataPoint log file in JSON or JSONL format (pandas orient=records)",
        type=lambda x: get_valid_path(parser, x),
    )
    parser.add_argument(
        "mode",
        help="Mode of operation for the sender. "
        "Either cherry-pick datapoints by attribute, or send the entire log.",
        choices=["cherry-pick", "all"],
    )
    parser.add_argument(
        "--endpoint-url",
        help="base DP3 endpoint, default='http://127.0.0.1:5000'",
        default="http://127.0.0.1:5000",
        type=str,
        dest="endpoint_url",
    )
    parser.add_argument(
        "-n",
        dest="count",
        help="number of DataPoints to send per attribute, default=50",
        default=50,
        type=int,
    )
    parser.add_argument(
        "-c",
        "--chunk",
        dest="chunk",
        help="number of DataPoints to send per request, default=same as count",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--shift-time",
        help="shift timestamps so that 't1' is the current UTC time",
        default=False,
        action="store_true",
        dest="shift_time",
    )
    parser.add_argument(
        "--workers",
        "-w",
        help="number of workers to use, default=1",
        default=1,
        type=int,
    )
    args = parser.parse_args()
    if args.chunk is None:
        args.chunk = args.count
    dp_factory = get_shifted_datapoint_from_row if args.shift_time else get_datapoint_from_row
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)-15s,%(threadName)s,[%(levelname)s] %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )
    log = logging.getLogger("DataPointSender")

    prev = 0
    dp_queue = Queue()
    telem_queue = Queue()

    # Send them
    queue_empty = Event()
    stop = Event()
    workers = [
        Thread(
            target=worker_thread,
            name=f"Worker-{i}",
            args=(i, args, stop, dp_queue, queue_empty, telem_queue),
        )
        for i in range(args.workers)
    ]
    reader = Thread(
        target=reader_thread,
        name="Reader",
        args=(args.input_file, dp_queue, queue_empty, stop),
        kwargs={"prefetch_size": max(4, args.workers * 2)},
    )
    try:
        reader.start()
        log.info("Starting %s workers", len(workers))
        [w.start() for w in workers]
        log.debug("Avaiting result")
        [w.join() for w in workers]
    except KeyboardInterrupt:
        log.info("Received user interrupt, exiting.")
        stop.set()
        queue_empty.set()
        reader.join()
        [w.join() for w in workers]

    dps_sent, request_time = 0, 0

    w_times = {i: 0 for i in range(args.workers)}
    while not telem_queue.empty():
        i, dps_sent_, request_time_ = telem_queue.get()
        dps_sent += dps_sent_
        w_times[i] += request_time_
    request_time = max(w_times.values())

    if dps_sent != 0:
        log.info(
            "Total: Sent %s datapoints in %.3fs (throughput: %.3f DPs/s; %.4f s/DP)",
            dps_sent,
            request_time,
            dps_sent / request_time,
            request_time / dps_sent,
        )
    else:
        log.info("No datapoints were sent.")



import logging
from functools import partial, update_wrapper

from dp3.common.base_module import BaseModule
from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import PlatformConfig


def copy_linked(_: str, record: dict, link: str, attr: str):
    record[attr] = record.get(link, {}).get(attr, None)


def modify_value(_: str, record: dict, attr: str, value):
    record[attr] = value


dummy_hook_abc = update_wrapper(partial(modify_value, attr="data2", value="abc"), modify_value)
dummy_hook_def = update_wrapper(partial(modify_value, attr="data1", value="def"), modify_value)


class TestModule(BaseModule):
    def __init__(
        self, platform_config: PlatformConfig, module_config: dict, registrar: CallbackRegistrar
    ):
        self.log = logging.getLogger("TestModule")
        self.log.setLevel("DEBUG")
        self.model_spec = platform_config.model_spec

        # just for testing purposes - as new value for test_attrib
        self.counter = module_config.get("init_value", 0)

        registrar.register_correlation_hook(
            update_wrapper(partial(copy_linked, link="bs", attr="data1"), copy_linked),
            "A",
            depends_on=[["bs", "data1"]],
            may_change=[["data1"]],
        )
        registrar.register_correlation_hook(
            update_wrapper(partial(copy_linked, link="as", attr="data2"), copy_linked),
            "B",
            depends_on=[["as", "data2"]],
            may_change=[["data2"]],
        )

        # Testing hook dependencies
        registrar.register_correlation_hook(
            update_wrapper(partial(copy_linked, link="ds", attr="data1"), copy_linked),
            "C",
            depends_on=[["ds", "data1"]],
            may_change=[["data1"]],
        )
        registrar.register_correlation_hook(
            update_wrapper(partial(copy_linked, link="cs", attr="data2"), copy_linked),
            "D",
            depends_on=[["cs", "data2"]],
            may_change=[["data2"]],
        )
        registrar.register_correlation_hook(
            update_wrapper(partial(modify_value, attr="data2", value="modifc"), modify_value),
            "C",
            depends_on=[],
            may_change=[["data2"]],
        )
        registrar.register_correlation_hook(
            update_wrapper(partial(modify_value, attr="data1", value="modifd"), modify_value),
            "D",
            depends_on=[],
            may_change=[["data1"]],
        )



from dp3.api.internal.config import MODEL_SPEC
from dp3.common.datapoint import DataPointBase


def api_to_dp3_datapoint(api_dp_values: dict) -> DataPointBase:
    """Converts API datapoint values to DP3 datapoint

    If etype-attr pair doesn't exist in DP3 config, raises `ValueError`.
    If values are not valid, raises pydantic's ValidationError.
    """
    etype = api_dp_values["type"]
    attr = api_dp_values["attr"]

    # Convert to DP3 datapoint format
    dp3_dp_values = api_dp_values
    dp3_dp_values["etype"] = etype
    dp3_dp_values["eid"] = api_dp_values["id"]

    # Get attribute-specific model
    try:
        model = MODEL_SPEC.attr(etype, attr).dp_model
    except KeyError as e:
        raise ValueError(f"Combination of type '{etype}' and attr '{attr}' doesn't exist") from e

    # Parse using the model
    # This may raise pydantic's ValidationError, but that's intensional (to get
    # a JSON-serializable trace as a response from API).
    return model.model_validate(dp3_dp_values)



import inspect
import logging
import re
import threading
import time
import urllib
from collections import defaultdict
from collections.abc import Iterable, Iterator
from datetime import datetime, timedelta
from typing import Any, Callable, Literal, Optional, Union

import bson
import pymongo
from event_count_logger import DummyEventGroup
from pydantic import BaseModel, Field, field_validator
from pymongo import ReplaceOne, UpdateMany, UpdateOne, WriteConcern
from pymongo.command_cursor import CommandCursor
from pymongo.cursor import Cursor
from pymongo.errors import BulkWriteError, DocumentTooLarge, OperationFailure, WriteError
from pymongo.results import DeleteResult, UpdateResult

from dp3.common.attrspec import AttrType, timeseries_types
from dp3.common.config import HierarchicalDict, ModelSpec
from dp3.common.datapoint import DataPointBase, to_json_friendly
from dp3.common.scheduler import Scheduler
from dp3.common.task import HASH
from dp3.common.types import EventGroupType
from dp3.database.schema_cleaner import SchemaCleaner

BSON_OBJECT_TOO_LARGE = 10334


class DatabaseError(Exception):
    pass


class MissingTableError(DatabaseError):
    pass


class MongoHostConfig(BaseModel, extra="forbid"):
    """MongoDB host."""

    address: str = "localhost"
    port: int = 27017


class MongoStandaloneConfig(BaseModel, extra="forbid"):
    """MongoDB standalone configuration."""

    mode: Literal["standalone"]
    host: MongoHostConfig = MongoHostConfig()


class MongoReplicaConfig(BaseModel, extra="forbid"):
    """MongoDB replica set configuration."""

    mode: Literal["replica"]
    replica_set: str = "dp3"
    hosts: list[MongoHostConfig]


class StorageConfig(BaseModel, extra="forbid"):
    """Storage configuration."""

    snapshot_bucket_size: int = 32


class MongoConfig(BaseModel, extra="forbid"):
    """Database configuration."""

    db_name: str = "dp3"
    username: str = "dp3"
    password: str = "dp3"
    connection: Union[MongoStandaloneConfig, MongoReplicaConfig] = Field(..., discriminator="mode")
    storage: StorageConfig = StorageConfig()

    @field_validator("username", "password")
    @classmethod
    def url_safety(cls, v):
        return urllib.parse.quote_plus(v)


def get_caller_id():
    """Returns the name of the caller method's class, or function name if caller is not a method."""
    caller = inspect.stack()[2]
    if module := caller.frame.f_locals.get("self"):
        return module.__class__.__qualname__
    return caller.function


# number of seconds to wait for the i-th attempt to reconnect after error
RECONNECT_DELAYS = [1, 2, 5, 10, 30]


class EntityDatabase:
    """
    MongoDB database wrapper responsible for whole communication with database server.
    Initializes database schema based on database configuration.

    Args:
        db_conf: configuration of database connection (content of database.yml)
        model_spec: ModelSpec object, configuration of data model (entities and attributes)
        process_index: index of worker process - used for sharding metadata
        num_processes: number of worker processes
    """

    def __init__(
        self,
        config: HierarchicalDict,
        model_spec: ModelSpec,
        num_processes: int,
        process_index: int = 0,
        elog: Optional[EventGroupType] = None,
    ) -> None:
        self.log = logging.getLogger("EntityDatabase")
        self.elog = elog or DummyEventGroup()

        db_config = MongoConfig.model_validate(config.get("database", {}))

        self.log.info("Connecting to database...")
        for attempt, delay in enumerate(RECONNECT_DELAYS):
            try:
                self._db = self.connect(db_config)
                # Check if connected
                self._db.admin.command("ping")
            except pymongo.errors.ConnectionFailure as e:
                if attempt + 1 == len(RECONNECT_DELAYS):
                    raise DatabaseError(
                        "Cannot connect to database with specified connection arguments."
                    ) from e
                else:
                    self.log.error(
                        "Cannot connect to database (attempt %d, retrying in %ds).",
                        attempt + 1,
                        delay,
                    )
                    time.sleep(delay)

        self._db_schema_config = model_spec
        self._num_processes = num_processes
        self._process_index = process_index

        self._normal_snapshot_eids = defaultdict(set)
        self._oversized_snapshot_eids = defaultdict(set)
        self._snapshot_bucket_size = db_config.storage.snapshot_bucket_size
        self._bucket_delta = self._get_snapshot_bucket_delta(config)

        # Init and switch to correct database
        self._db = self._db[db_config.db_name]
        type_registry = bson.codec_options.TypeRegistry(fallback_encoder=to_json_friendly)
        self._codec_opts = bson.codec_options.CodecOptions(type_registry=type_registry)

        if process_index == 0:
            self._init_database_schema(db_config.db_name)

        self.schema_cleaner = SchemaCleaner(
            self._db, self.get_module_cache("Schema"), self._db_schema_config, config, self.log
        )

        self._on_entity_delete_one = []
        self._on_entity_delete_many = []

        self._raw_buffer_locks = {etype: threading.Lock() for etype in model_spec.entities}
        self._raw_buffers = defaultdict(list)
        self._master_buffer_locks = {etype: threading.Lock() for etype in model_spec.entities}
        self._master_buffers = defaultdict(dict)

        self._sched = Scheduler()
        seconds = ",".join(
            f"{int(i)}" for i in range(60) if int(i - process_index) % min(num_processes, 3) == 0
        )
        self._sched.register(self._push_raw, second=seconds, misfire_grace_time=5)
        self._sched.register(self._push_master, second=seconds, misfire_grace_time=5)

        self.log.info("Database successfully initialized!")

    def _get_snapshot_bucket_delta(self, config) -> timedelta:
        """Returns how long it takes to fill a snapshot bucket.

        This depends on the frequency of snapshot creation and the bucket size.
        """
        creation_rate = config.get("snapshots.creation_rate", {"minute": "*/30"})
        if len(creation_rate) > 1:
            raise ValueError("Only one snapshot creation rate is supported.")

        time_to_sec = {
            "second": 1,
            "minute": 60,
            "hour": 60 * 60,
            "day": 60 * 60 * 24,
        }

        time_to_sec_shifted = {
            "second": 60,
            "minute": 60 * 60,
            "hour": 60 * 60 * 24,
            "day": 60 * 60 * 24 * 30,
        }

        for key, value in creation_rate.items():
            if value.startswith("*/"):
                seconds = time_to_sec.get(key)
                if seconds is None:
                    raise ValueError(f"Unsupported snapshot creation rate: {creation_rate}")

                snapshot_interval = int(value[2:])
                bucket_delta = seconds * snapshot_interval
            else:
                seconds = time_to_sec_shifted.get(key)
                if seconds is None:
                    raise ValueError(f"Unsupported snapshot creation rate: {creation_rate}")

                count = len(value.split(","))
                bucket_delta = seconds // count
            break
        else:
            raise ValueError(f"Unsupported snapshot creation rate: {creation_rate}")

        return timedelta(seconds=bucket_delta * self._snapshot_bucket_size)

    @staticmethod
    def connect(config: MongoConfig) -> pymongo.MongoClient:
        if isinstance(config.connection, MongoStandaloneConfig):
            host = config.connection.host
            return pymongo.MongoClient(
                f"mongodb://{config.username}:{config.password}@{host.address}:{host.port}/",
                connectTimeoutMS=3000,
                serverSelectionTimeoutMS=5000,
            )
        elif isinstance(config.connection, MongoReplicaConfig):
            uri = (
                f"mongodb://{config.username}:{config.password}@"
                + ",".join(f"{host.address}:{host.port}" for host in config.connection.hosts)
                + f"/?replicaSet={config.connection.replica_set}"
            )
            return pymongo.MongoClient(
                uri,
                replicaSet=config.connection.replica_set,
                connectTimeoutMS=3000,
                serverSelectionTimeoutMS=5000,
            )
        else:
            raise NotImplementedError()

    def start(self) -> None:
        """Starts the database sync of raw datapoint inserts."""
        self._sched.start()

    def stop(self) -> None:
        """Stops the database sync, push remaining datapoints."""
        self._sched.stop()
        self._push_raw()
        self._push_master()

    def register_on_entity_delete(
        self, f_one: Callable[[str, str], None], f_many: Callable[[str, list[str]], None]
    ):
        """Registers function to be called when entity is forcibly deleted."""
        self._on_entity_delete_one.append(f_one)
        self._on_entity_delete_many.append(f_many)

    @staticmethod
    def _master_col_name(entity: str) -> str:
        """Returns name of master collection for `entity`."""
        return f"{entity}#master"

    @staticmethod
    def _snapshots_col_name(entity: str) -> str:
        """Returns name of snapshots collection for `entity`."""
        return f"{entity}#snapshots"

    @staticmethod
    def _oversized_snapshots_col_name(entity: str) -> str:
        """Returns name of oversized snapshots collection for `entity`."""
        return f"{entity}#snapshots_oversized"

    @staticmethod
    def _raw_col_name(entity: str) -> str:
        """Returns name of raw data collection for `entity`."""
        return f"{entity}#raw"

    @staticmethod
    def _get_new_archive_col_name(entity: str) -> str:
        """Returns name of new archive collection for `entity`."""
        return f"{entity}#archive_{datetime.utcnow().strftime('%Y_%m_%d_%H%M%S')}"

    def _archive_col_names(self, entity: str) -> list[str]:
        """Returns names of archive collections for `entity`."""
        return self._db.list_collection_names(filter={"name": {"$regex": f"{entity}#archive.*"}})

    def _init_database_schema(self, db_name) -> None:
        """Runs full check and update of database schema.

        Checks whether `db_name` database exists. If not, creates it.
        Checks whether all collections exist. If not, creates them.
        This all is done automatically in MongoDB.

        We need to create indexes, however.
        """
        for etype in self._db_schema_config.entities:
            # Snapshots index on `eid` and `_time_created` fields
            snapshot_col = self._oversized_snapshots_col_name(etype)
            self._db[snapshot_col].create_index("eid", background=True)
            self._db[snapshot_col].create_index("_time_created", background=True)

            snapshot_col = self._snapshots_col_name(etype)
            # To get the empty bucket for an entity without fetching all its buckets
            for ix in self._db[snapshot_col].list_indexes():
                if set(ix["key"].keys()) == {"_id", "count"}:
                    if ix.get("partialFilterExpression") is None:
                        self._db[snapshot_col].drop_index(ix["name"])
                        self._db[snapshot_col].create_index(
                            [("_id", pymongo.DESCENDING), ("count", pymongo.ASCENDING)],
                            partialFilterExpression={"count": {"$lt": self._snapshot_bucket_size}},
                            background=True,
                        )
                    break
            else:
                self._db[snapshot_col].create_index(
                    [("_id", pymongo.DESCENDING), ("count", pymongo.ASCENDING)],
                    partialFilterExpression={"count": {"$lt": self._snapshot_bucket_size}},
                    background=True,
                )

            # To fetch all the latest snapshots without relying on date
            self._db[snapshot_col].create_index(
                [("latest", pymongo.DESCENDING), ("_id", pymongo.ASCENDING)],
                partialFilterExpression={"latest": {"$eq": True}},
                background=True,
            )

            # To fetch the oversized entities only
            self._db[snapshot_col].create_index(
                [("oversized", pymongo.DESCENDING)],
                partialFilterExpression={"oversized": {"$eq": True}},
                background=True,
            )
            # For deleting old snapshots
            self._db[snapshot_col].create_index("_time_created", background=True)

        # Create a TTL index for metadata collection
        self._db["#metadata"].create_index(
            "#time_created", expireAfterSeconds=60 * 60 * 24 * 30, background=True
        )

        # Master indexes
        for etype in self._db_schema_config.entities:
            master_col = self._db[self._master_col_name(etype)]
            history_attrs = set()
            plain_attrs = set()

            for attr, spec in self._db_schema_config.entity_attributes[etype].items():
                if spec.t == AttrType.PLAIN:
                    plain_attrs.add(attr)
                if spec.t in AttrType.TIMESERIES | AttrType.OBSERVATIONS:
                    history_attrs.add(attr)

            attrs = history_attrs | plain_attrs

            # Drop previous indexes
            create_wildcard = True
            for index in master_col.list_indexes():
                if index["name"] == "_id_":
                    continue
                attr_names = set(index["key"].keys())

                # Drop individual attribute indexes
                if attr_names & attrs:
                    self.log.info("Dropping index %s on %s", index["name"], etype)
                    master_col.drop_index(index["name"])

                # Drop wildcard indexes if there are missing attributes
                if "$**" in attr_names:
                    if index.get("wildcardProjection") is not None:
                        covered_attrs = {
                            it
                            for k in index["wildcardProjection"]
                            for it in k.rsplit(".", maxsplit=1)
                        }
                        covered_attrs -= {"ts_last_update", "#min_t2s"}
                        if "t2" not in covered_attrs and not attrs - covered_attrs:
                            create_wildcard = False
                            continue  # Index already covers all history attributes
                    self.log.info("Dropping wildcard index %s on %s", index["name"], etype)
                    master_col.drop_index(index["name"])

            # Create an index on the id hash
            master_col.create_index("#hash", background=True)

            if not create_wildcard:
                continue

            # Create wildcard index for attribute histories
            index_name = f"wildcard_{datetime.now().strftime('%Y%m%d%H%M%S')}"
            try:
                master_col.create_index(
                    [("$**", 1)],
                    name=index_name,
                    wildcardProjection={f"#min_t2s.{attr}": 1 for attr in history_attrs}
                    | {f"{attr}.ts_last_update": 1 for attr in plain_attrs},
                )
                self.log.info("Created wildcard index %s on %s", index_name, etype)
            except OperationFailure as why:
                self.log.error("Failed to create wildcard index: %s", why)

    def _assert_etype_exists(self, etype: str):
        """Asserts `etype` existence.

        If doesn't exist, raises `DatabaseError`.
        """
        if etype not in self._db_schema_config.entities:
            raise DatabaseError(f"Entity '{etype}' does not exist")

    def update_schema(self):
        """
        Checks whether schema saved in database is up-to-date and updates it if necessary.

        Will NOT perform any changes in master collections on conflicting model changes.
        Any such changes must be performed via the CLI.

        As this method modifies the schema collection, it should be called only by the main worker.
        """
        try:
            self.schema_cleaner.safe_update_schema()
        except ValueError as e:
            raise DatabaseError("Schema update failed. Please run `dp3 schema-update`.") from e

    def await_updated_schema(self):
        """
        Checks whether schema saved in database is up-to-date and awaits its update
        by the main worker on mismatch.
        """
        self.schema_cleaner.await_updated()

    def insert_datapoints(
        self, etype: str, eid: str, dps: list[DataPointBase], new_entity: bool = False
    ) -> None:
        """Inserts datapoint to raw data collection and updates master record.

        Raises DatabaseError when insert or update fails.
        """
        if len(dps) == 0:
            return

        etype = dps[0].etype

        # Check `etype`
        self._assert_etype_exists(etype)

        # Insert raw datapoints
        dps_dicts = [dp.model_dump(exclude={"attr_type"}) for dp in dps]
        with self._raw_buffer_locks[etype]:
            self._raw_buffers[etype].extend(dps_dicts)

        # Update master document
        master_changes = {"pushes": defaultdict(list), "$set": {}}
        for dp in dps:
            attr_spec = self._db_schema_config.attr(etype, dp.attr)

            if attr_spec.t in AttrType.PLAIN | AttrType.OBSERVATIONS and attr_spec.is_iterable:
                v = [elem.model_dump() if isinstance(elem, BaseModel) else elem for elem in dp.v]
            else:
                v = dp.v.model_dump() if isinstance(dp.v, BaseModel) else dp.v

            # Rewrite value of plain attribute
            if attr_spec.t == AttrType.PLAIN:
                master_changes["$set"][dp.attr] = {"v": v, "ts_last_update": datetime.now()}

            # Push new data of observation
            if attr_spec.t == AttrType.OBSERVATIONS:
                master_changes["pushes"][dp.attr].append(
                    {"t1": dp.t1, "t2": dp.t2, "v": v, "c": dp.c}
                )

            # Push new data of timeseries
            if attr_spec.t == AttrType.TIMESERIES:
                master_changes["pushes"][dp.attr].append({"t1": dp.t1, "t2": dp.t2, "v": v})

        if new_entity:
            master_changes["$set"]["#hash"] = HASH(f"{etype}:{eid}")
            master_changes["$set"]["#time_created"] = datetime.now()

        with self._master_buffer_locks[etype]:
            if eid in self._master_buffers[etype]:
                for attr, push_dps in master_changes["pushes"].items():
                    if attr in self._master_buffers[etype][eid]["pushes"]:
                        self._master_buffers[etype][eid]["pushes"][attr].extend(push_dps)
                    else:
                        self._master_buffers[etype][eid]["pushes"][attr] = push_dps
                self._master_buffers[etype][eid]["$set"].update(master_changes["$set"])
            else:
                self._master_buffers[etype][eid] = master_changes

    def _push_raw(self):
        """Pushes datapoints to raw collections."""
        for etype, lock in self._raw_buffer_locks.items():
            begin = time.time()
            with lock:
                locked = time.time()
                dps = self._raw_buffers[etype]
                self._raw_buffers[etype] = []
                if not dps:
                    continue
            try:
                self._db.get_collection(
                    self._raw_col_name(etype),
                    write_concern=WriteConcern(w=0),
                    codec_options=self._codec_opts,
                ).insert_many(dps, ordered=False)
                end = time.time()
                self.log.debug(
                    "Inserted %s raw datapoints to %s in %.3fs, %.3fus lock, %.3fs insert",
                    len(dps),
                    etype,
                    end - begin,
                    (locked - begin) * 1000_000,
                    end - locked,
                )
            except Exception as e:
                raise DatabaseError(f"Insert of datapoints failed: {e}\n{dps}") from e

    def _push_master(self):
        """Push master changes to database."""
        for etype, lock in self._master_buffer_locks.items():
            master_col = self._db.get_collection(
                self._master_col_name(etype),
                write_concern=WriteConcern(w=1),
                codec_options=self._codec_opts,
            )
            begin = time.time()
            with lock:
                locked = time.time()
                master_changes = self._master_buffers[etype]
                self._master_buffers[etype] = {}
            if not master_changes:
                continue
            try:
                updates = [
                    UpdateOne(
                        {"_id": eid},
                        {
                            "$push": {
                                attr: push_dps[0] if len(push_dps) == 1 else {"$each": push_dps}
                                for attr, push_dps in changes["pushes"].items()
                            },
                            "$set": changes["$set"],
                        },
                        upsert=True,
                    )
                    for eid, changes in master_changes.items()
                ]
                updates_ready = time.time()
                res = master_col.bulk_write(updates, ordered=False)
                end = time.time()
                self.log.debug(
                    "Updated %s master records in %.3fs, "
                    "%.3fus lock, %.3fms update prep, %.3fs update",
                    len(master_changes),
                    end - begin,
                    (locked - begin) * 1000_000,
                    (updates_ready - locked) * 1000,
                    end - updates_ready,
                )
                for error in res.bulk_api_result.get("writeErrors", []):
                    self.log.error("Error in bulk write: %s", error)
            except Exception as e:
                raise DatabaseError(
                    f"Update of master records failed: {e}\n{master_changes}"
                ) from e

    def update_master_records(self, etype: str, eids: list[str], records: list[dict]) -> None:
        """Replace master records of `etype`:`eid` with the provided `records`.

        Raises DatabaseError when update fails.
        """
        master_col = self._master_col_name(etype)
        try:
            res = self._db[master_col].bulk_write(
                [
                    ReplaceOne({"_id": eid}, record, upsert=True)
                    for eid, record in zip(eids, records)
                ],
                ordered=False,
            )
            self.log.debug("Updated master records of %s (%s).", etype, len(eids))
            for error in res.bulk_api_result.get("writeErrors", []):
                self.log.error("Error in bulk write: %s", error)
        except Exception as e:
            raise DatabaseError(f"Update of master records failed: {e}\n{records}") from e

    def extend_ttl(self, etype: str, eid: str, ttl_tokens: dict[str, datetime]):
        """Extends TTL of given `etype`:`eid` by `ttl_tokens`."""
        master_col = self._master_col_name(etype)
        try:
            self._db[master_col].update_one(
                {"_id": eid},
                {
                    "$max": {
                        f"#ttl.{token_name}": token_value
                        for token_name, token_value in ttl_tokens.items()
                    }
                },
            )
            self.log.debug("Updated TTL of %s: %s.", etype, eid)
        except Exception as e:
            raise DatabaseError(f"TTL update failed: {e} ({ttl_tokens})") from e

    def remove_expired_ttls(self, etype: str, expired_eid_ttls: dict[str, list[str]]):
        """Removes expired TTL of given `etype`:`eid`."""
        master_col = self._master_col_name(etype)
        try:
            res = self._db[master_col].bulk_write(
                [
                    UpdateOne(
                        {"_id": eid},
                        {"$unset": {f"#ttl.{token_name}": "" for token_name in expired_ttls}},
                    )
                    for eid, expired_ttls in expired_eid_ttls.items()
                ]
            )
            self.log.debug(
                "Removed expired TTL of %s: (%s, modified %s).",
                etype,
                len(expired_eid_ttls),
                res.modified_count,
            )
        except Exception as e:
            raise DatabaseError(f"TTL update failed: {e}") from e

    def delete_eids(self, etype: str, eids: list[str]):
        """Delete master record and all snapshots of `etype`:`eids`."""
        master_col = self._master_col_name(etype)
        snapshot_col = self._snapshots_col_name(etype)
        os_snapshot_col = self._oversized_snapshots_col_name(etype)
        try:
            res = self._db[master_col].delete_many({"_id": {"$in": eids}})
            self.log.debug(
                "Deleted %s master records of %s (%s).", res.deleted_count, etype, len(eids)
            )
            self.elog.log("record_removed", count=res.deleted_count)
        except Exception as e:
            raise DatabaseError(f"Delete of master record failed: {e}\n{eids}") from e
        try:
            res = self._db[snapshot_col].delete_many(self._snapshot_bucket_eids_filter(eids))
            del_cnt = res.deleted_count * self._snapshot_bucket_size
            self.log.debug("Deleted %s snapshots of %s (%s).", del_cnt, etype, len(eids))
            res = self._db[os_snapshot_col].delete_many({"eid": {"$in": eids}})
            self.log.debug(
                "Deleted %s oversized snapshots of %s (%s).", res.deleted_count, etype, len(eids)
            )
        except Exception as e:
            raise DatabaseError(f"Delete of snapshots failed: {e}\n{eids}") from e
        for f in self._on_entity_delete_many:
            try:
                f(etype, eids)
            except Exception as e:
                self.log.exception("Error in on_entity_delete_many callback %s: %s", f, e)

    def delete_eid(self, etype: str, eid: str):
        """Delete master record and all snapshots of `etype`:`eid`."""
        master_col = self._master_col_name(etype)
        snapshot_col = self._snapshots_col_name(etype)
        os_snapshot_col = self._oversized_snapshots_col_name(etype)
        try:
            self._db[master_col].delete_one({"_id": eid})
            self.log.debug("Deleted master record of %s/%s.", etype, eid)
            self.elog.log("record_removed")
        except Exception as e:
            raise DatabaseError(f"Delete of master record failed: {e}\n{eid}") from e
        try:
            res = self._db[snapshot_col].delete_many(self._snapshot_bucket_eid_filter(eid))
            del_cnt = res.deleted_count * self._snapshot_bucket_size
            self.log.debug("deleted %s snapshots of %s/%s.", del_cnt, etype, eid)
            res = self._db[os_snapshot_col].delete_many({"eid": eid})
            self.log.debug(
                "Deleted %s oversized snapshots of %s/%s.", res.deleted_count, etype, eid
            )
        except Exception as e:
            raise DatabaseError(f"Delete of snapshots failed: {e}\n{eid}") from e
        for f in self._on_entity_delete_one:
            try:
                f(etype, eid)
            except Exception as e:
                self.log.exception("Error in on_entity_delete_one callback %s: %s", f, e)

    def mark_all_entity_dps_t2(self, etype: str, attrs: list[str]) -> UpdateResult:
        """
        Updates the `min_t2s` of the master records of `etype` for all records.

        Periodically called for all `etype`s from HistoryManager.
        """
        master_col = self._master_col_name(etype)
        try:
            return self._db[master_col].update_many(
                {},
                [
                    {
                        "$set": {
                            attr_name: {
                                "$cond": {
                                    "if": {
                                        "$eq": [
                                            {"$size": {"$ifNull": [f"${attr_name}", []]}},
                                            0,
                                        ]
                                    },
                                    "then": "$$REMOVE",
                                    "else": f"${attr_name}",
                                }
                            }
                            for attr_name in attrs
                        }
                        | {
                            f"#min_t2s.{attr_name}": {
                                "$cond": {
                                    "if": {
                                        "$eq": [
                                            {"$size": {"$ifNull": [f"${attr_name}", []]}},
                                            0,
                                        ]
                                    },
                                    "then": "$$REMOVE",
                                    "else": {"$min": f"${attr_name}.t2"},
                                }
                            }
                            for attr_name in attrs
                        }
                    }
                ],
            )
        except Exception as e:
            raise DatabaseError(f"Update of min_t2s failed: {e}") from e

    def delete_old_dps(self, etype: str, attr_name: str, t_old: datetime) -> UpdateResult:
        """Delete old datapoints from master collection.

        Periodically called for all `etype`s from HistoryManager.
        """
        master_col = self._db.get_collection(
            self._master_col_name(etype), write_concern=WriteConcern(w=1)
        )
        try:
            return master_col.update_many(
                {f"#min_t2s.{attr_name}": {"$lt": t_old}},
                [
                    {
                        "$set": {
                            attr_name: {
                                "$filter": {
                                    "input": f"${attr_name}",
                                    "cond": {"$gte": ["$$this.t2", t_old]},
                                }
                            }
                        }
                    },
                    {
                        "$set": {
                            f"#min_t2s.{attr_name}": {
                                "$cond": {
                                    "if": {
                                        "$eq": [
                                            {"$size": {"$ifNull": [f"${attr_name}", []]}},
                                            0,
                                        ]
                                    },
                                    "then": "$$REMOVE",
                                    "else": {"$min": f"${attr_name}.t2"},
                                }
                            }
                        },
                    },
                ],
            )
        except Exception as e:
            raise DatabaseError(f"Delete of old datapoints failed: {e}") from e

    def delete_link_dps(
        self, etype: str, affected_eids: list[str], attr_name: str, eid_to: str
    ) -> None:
        """Delete link datapoints from master collection.

        Called from LinkManager for deleted entities.
        """
        master_col = self._master_col_name(etype)
        attr_type = self._db_schema_config.attr(etype, attr_name).t
        filter_cond = {"_id": {"$in": affected_eids}}
        try:
            if attr_type == AttrType.OBSERVATIONS:
                update_pull = {"$pull": {attr_name: {"v.eid": eid_to}}}
                self._db[master_col].update_many(filter_cond, update_pull)
            elif attr_type == AttrType.PLAIN:
                update_unset = {"$unset": {attr_name: ""}}
                self._db[master_col].update_many(filter_cond, update_unset)
            else:
                raise ValueError(f"Unsupported attribute type: {attr_type}")
        except Exception as e:
            raise DatabaseError(f"Delete of link datapoints failed: {e}") from e

    def delete_many_link_dps(
        self,
        etypes: list[str],
        affected_eids: list[list[str]],
        attr_names: list[str],
        eids_to: list[list[str]],
    ) -> None:
        """Delete link datapoints from master collection.

        Called from LinkManager for deleted entities, when deleting multiple entities.
        """
        try:
            updates = []
            for etype, affected_eid_list, attr_name, eid_to_list in zip(
                etypes, affected_eids, attr_names, eids_to
            ):
                master_col = self._master_col_name(etype)
                attr_type = self._db_schema_config.attr(etype, attr_name).t
                filter_cond = {"_id": {"$in": affected_eid_list}}
                if attr_type == AttrType.OBSERVATIONS:
                    update_pull = {"$pull": {attr_name: {"v.eid": {"$in": eid_to_list}}}}
                    updates.append(UpdateMany(filter_cond, update_pull))
                elif attr_type == AttrType.PLAIN:
                    update_unset = {"$unset": {attr_name: ""}}
                    updates.append(UpdateMany(filter_cond, update_unset))
                else:
                    raise ValueError(f"Unsupported attribute type: {attr_type}")
                self._db[master_col].bulk_write(updates)
        except Exception as e:
            raise DatabaseError(f"Delete of link datapoints failed: {e}") from e

    def get_master_record(self, etype: str, eid: str, **kwargs) -> dict:
        """Get current master record for etype/eid.

        If doesn't exist, returns {}.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        master_col = self._master_col_name(etype)
        return self._db[master_col].find_one({"_id": eid}, **kwargs) or {}

    def ekey_exists(self, etype: str, eid: str) -> bool:
        """Checks whether master record for etype/eid exists"""
        return bool(self.get_master_record(etype, eid))

    def get_master_records(self, etype: str, **kwargs) -> Cursor:
        """Get cursor to current master records of etype."""
        # Check `etype`
        self._assert_etype_exists(etype)

        master_col = self._master_col_name(etype)
        return self._db[master_col].find({}, **kwargs)

    def get_worker_master_records(
        self, worker_index: int, worker_cnt: int, etype: str, query_filter: dict = None, **kwargs
    ) -> Cursor:
        """Get cursor to current master records of etype."""
        if etype not in self._db_schema_config.entities:
            raise DatabaseError(f"Entity '{etype}' does not exist")

        query_filter = {} if query_filter is None else query_filter
        master_col = self._master_col_name(etype)
        return self._db[master_col].find(
            {"#hash": {"$mod": [worker_cnt, worker_index]}, **query_filter}, **kwargs
        )

    def get_latest_snapshot(self, etype: str, eid: str) -> dict:
        """Get latest snapshot of given etype/eid.

        If doesn't exist, returns {}.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        snapshot_col = self._snapshots_col_name(etype)
        return (
            self._db[snapshot_col].find_one(
                self._snapshot_bucket_eid_filter(eid), {"last": 1}, sort=[("_id", -1)]
            )
            or {}
        ).get("last", {})

    def _get_latest_snapshots_date(self) -> Optional[datetime]:
        """Get date of newest snapshot set.

        Queries snapshots metadata to find latest fully completed set of snapshots
        and returns it's date.
        If it doesn't exist, returns `None`.

        Note: This doesn't take into account etypes, so it may be inaccurate for new entity types.
        """
        # Get latest fully completed snapshot metadata
        lfcsm = self._db["#metadata"].find_one(
            {
                "#module": "SnapShooter",
                "workers_finished": self._num_processes,
                "linked_finished": True,
            },
            sort=[("#time_created", -1)],
        )

        if lfcsm is None:
            return None

        # Extract date
        return lfcsm["#time_created"]

    def get_latest_snapshots(
        self,
        etype: str,
        fulltext_filters: Optional[dict[str, str]] = None,
        generic_filter: Optional[dict[str, any]] = None,
    ) -> tuple[Cursor, int]:
        """Get latest snapshots of given `etype`.

        This method is useful for displaying data on web.

        Returns only documents matching `generic_filter` and `fulltext_filters`
        (dictionary attribute - fulltext filter).
        Fulltext filters are interpreted as regular expressions.
        Only string values may be filtered this way. There's no validation that queried attribute
        can be fulltext filtered.
        Only plain and observation attributes with string-based data types can be queried.
        Array and set data types are supported as well as long as they are not multi value
        at the same time.
        If you need to filter EIDs, use attribute `eid`.

        Generic filter allows filtering using generic MongoDB query (including `$and`, `$or`,
        `$lt`, etc.).
        There are no attribute name checks (may be added in the future).

        Generic and fulltext filters are merged - fulltext overrides conflicting keys.

        Also returns total document count (after filtering).

        May raise `DatabaseError` if query is invalid.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        snapshot_col = self._snapshots_col_name(etype)

        if not fulltext_filters:
            fulltext_filters = {}

        if not generic_filter:
            generic_filter = {}

        # Create base of query
        query = generic_filter
        query["latest"] = True

        # Process fulltext filters
        for attr in fulltext_filters:
            fulltext_filter = {"$regex": fulltext_filters[attr], "$options": "i"}

            # EID filter
            if attr == "eid":
                query["_id"] = fulltext_filter
                continue

            # Check if attribute exists
            try:
                attr_spec = self._db_schema_config.attr(etype, attr)
            except KeyError as e:
                raise DatabaseError(f"Attribute '{attr}' in fulltext filter doesn't exist") from e

            # Correctly handle link<...> data type
            if attr_spec.t in AttrType.PLAIN | AttrType.OBSERVATIONS and attr_spec.is_relation:
                query["last." + attr + ".eid"] = fulltext_filter
            else:
                query["last." + attr] = fulltext_filter

        try:
            return self._db[snapshot_col].find(query, {"last": 1}).sort(
                [("_id", pymongo.ASCENDING)]
            ), self._db[snapshot_col].count_documents(query)
        except OperationFailure as e:
            raise DatabaseError("Invalid query") from e

    @staticmethod
    def _snapshot_bucket_id(eid: str, ctime: datetime) -> str:
        return f"{eid}_#{int(ctime.timestamp())}"

    @staticmethod
    def _snapshot_bucket_eid_filter(eid: str) -> dict:
        return {"_id": {"$regex": f"^{re.escape(eid)}_#"}}

    @staticmethod
    def _snapshot_bucket_eids_filter(eids: Iterable[str]) -> dict:
        return {"_id": {"$regex": "|".join([f"^{re.escape(eid)}_#" for eid in eids])}}

    def get_snapshots(
        self, etype: str, eid: str, t1: Optional[datetime] = None, t2: Optional[datetime] = None
    ) -> Union[Cursor, CommandCursor]:
        """Get all (or filtered) snapshots of given `eid`.

        This method is useful for displaying `eid`'s history on web.

        Args:
            etype: entity type
            eid: id of entity, to which data-points correspond
            t1: left value of time interval (inclusive)
            t2: right value of time interval (inclusive)
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        snapshot_col = self._snapshots_col_name(etype)

        # Find out if the snapshot is oversized
        doc = (
            self._db[snapshot_col]
            .find(self._snapshot_bucket_eid_filter(eid), {"oversized": 1})
            .sort([("_id", -1)])
            .limit(1)
        )
        doc = next(doc, None)
        if doc and doc.get("oversized", False):
            return self._get_snapshots_oversized(etype, eid, t1, t2)

        query = {"_time_created": {}}
        pipeline = [
            {"$match": self._snapshot_bucket_eid_filter(eid)},
            {"$unwind": "$history"},
            {"$replaceRoot": {"newRoot": "$history"}},
        ]

        # Filter by date
        if t1:
            query["_time_created"]["$gte"] = t1
        if t2:
            query["_time_created"]["$lte"] = t2

        # Unset if empty
        if query["_time_created"]:
            pipeline.append({"$match": query})
        pipeline.append({"$sort": {"_time_created": pymongo.ASCENDING}})
        return self._db[snapshot_col].aggregate(pipeline)

    def _get_snapshots_oversized(
        self, etype: str, eid: str, t1: Optional[datetime] = None, t2: Optional[datetime] = None
    ) -> Cursor:
        """Get all (or filtered) snapshots of given `eid` from oversized snapshots collection."""
        snapshot_col = self._oversized_snapshots_col_name(etype)
        query = {"eid": eid, "_time_created": {}}

        # Filter by date
        if t1:
            query["_time_created"]["$gte"] = t1
        if t2:
            query["_time_created"]["$lte"] = t2

        # Unset if empty
        if not query["_time_created"]:
            del query["_time_created"]

        return (
            self._db[snapshot_col]
            .find(query, projection={"_id": False})
            .sort([("_time_created", pymongo.ASCENDING)])
        )

    def get_value_or_history(
        self,
        etype: str,
        attr_name: str,
        eid: str,
        t1: Optional[datetime] = None,
        t2: Optional[datetime] = None,
    ) -> dict:
        """Gets current value and/or history of attribute for given `eid`.

        Depends on attribute type:
        - plain: just (current) value
        - observations: (current) value and history stored in master record (optionally filtered)
        - timeseries: just history stored in master record (optionally filtered)

        Returns dict with two keys: `current_value` and `history` (list of values).
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        attr_spec = self._db_schema_config.attr(etype, attr_name)

        result = {"current_value": None, "history": []}

        # Add current value to the result
        if attr_spec.t == AttrType.PLAIN:
            result["current_value"] = (
                self.get_master_record(etype, eid).get(attr_name, {}).get("v", None)
            )
        elif attr_spec.t == AttrType.OBSERVATIONS:
            result["current_value"] = self.get_latest_snapshot(etype, eid).get(attr_name, None)

        # Add history
        if attr_spec.t == AttrType.OBSERVATIONS:
            result["history"] = self.get_observation_history(etype, attr_name, eid, t1, t2)
        elif attr_spec.t == AttrType.TIMESERIES:
            result["history"] = self.get_timeseries_history(etype, attr_name, eid, t1, t2)

        return result

    def estimate_count_eids(self, etype: str) -> int:
        """Estimates count of `eid`s in given `etype`"""
        # Check `etype`
        self._assert_etype_exists(etype)

        master_col = self._master_col_name(etype)
        return self._db[master_col].estimated_document_count({})

    def _migrate_to_oversized_snapshot(self, etype: str, eid: str, snapshot: dict):
        snapshot_col = self._snapshots_col_name(etype)
        os_col = self._oversized_snapshots_col_name(etype)

        try:
            move_to_oversized = []
            last_id = None
            for doc in (
                self._db[snapshot_col].find(self._snapshot_bucket_eid_filter(eid)).sort({"_id": 1})
            ):
                move_to_oversized.extend(doc.get("history", []))
                last_id = doc["_id"]
            move_to_oversized.insert(0, snapshot)

            self._db[os_col].insert_many(move_to_oversized)
            self._db[snapshot_col].update_one(
                {"_id": last_id},
                {
                    "$set": {"oversized": True, "last": snapshot, "count": 0},
                    "$unset": {"history": ""},
                },
            )
            self._db[snapshot_col].delete_many(
                self._snapshot_bucket_eid_filter(eid) | {"oversized": False}
            )
        except Exception as e:
            raise DatabaseError(f"Update of snapshot {eid} failed: {e}, {snapshot}") from e

    def save_snapshot(self, etype: str, snapshot: dict, ctime: datetime):
        """Saves snapshot to specified entity of current master document.

        Will move snapshot to oversized snapshots if the maintained bucket is too large.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        if "eid" not in snapshot:
            self.log.error("Snapshot is missing 'eid' field: %s", snapshot)
            return
        eid = snapshot["eid"]
        snapshot["_time_created"] = ctime

        snapshot_col = self._snapshots_col_name(etype)
        os_col = self._oversized_snapshots_col_name(etype)

        # Find out if the snapshot is oversized
        normal, oversized = self._get_snapshot_state(etype, {eid})
        if normal:
            try:
                res = self._db[snapshot_col].update_one(
                    self._snapshot_bucket_eid_filter(eid)
                    | {"count": {"$lt": self._snapshot_bucket_size}},
                    {
                        "$set": {"last": snapshot},
                        "$push": {"history": {"$each": [snapshot], "$position": 0}},
                        "$inc": {"count": 1},
                        "$setOnInsert": {
                            "_id": self._snapshot_bucket_id(eid, ctime),
                            "_time_created": ctime,
                            "oversized": False,
                            "latest": True,
                        },
                    },
                    upsert=True,
                )

                if res.upserted_id is not None:
                    self._db[snapshot_col].update_many(
                        self._snapshot_bucket_eid_filter(eid)
                        | {"latest": True, "count": self._snapshot_bucket_size},
                        {"$unset": {"latest": 1}},
                    )
            except (WriteError, OperationFailure, DocumentTooLarge) as e:
                if e.code != BSON_OBJECT_TOO_LARGE:
                    raise e
                # The snapshot is too large, move it to oversized snapshots
                self.log.info(f"Snapshot of {eid} is too large: {e}, marking as oversized.")
                self._migrate_to_oversized_snapshot(etype, eid, snapshot)
                self._cache_snapshot_state(etype, set(), normal)
            except Exception as e:
                raise DatabaseError(f"Insert of snapshot {eid} failed: {e}, {snapshot}") from e
            return
        elif oversized:
            # Snapshot is already marked as oversized
            self._db[snapshot_col].update_one(
                self._snapshot_bucket_eid_filter(eid), {"$set": {"last": snapshot}}
            )
            self._db[os_col].insert_one(snapshot)
            return

    def _get_snapshot_state(self, etype: str, eids: set[str]) -> tuple[set, set]:
        """Get current state of snapshot of given `eid`."""
        unknown = eids
        normal = self._normal_snapshot_eids[etype] & unknown
        oversized = self._oversized_snapshot_eids[etype] & unknown
        unknown = unknown - normal - oversized

        if not unknown:
            return normal, oversized

        snapshot_col = self._snapshots_col_name(etype)
        new_oversized = set()
        for doc in self._db[snapshot_col].find(
            self._snapshot_bucket_eids_filter(unknown) | {"oversized": True},
            {"oversized": 1},
        ):
            eid = doc["_id"].rsplit("_#", maxsplit=1)[0]
            new_oversized.add(eid)

        unknown = unknown - new_oversized
        self._normal_snapshot_eids[etype] |= unknown
        self._oversized_snapshot_eids[etype] |= new_oversized

        return normal | unknown, oversized | new_oversized

    def _cache_snapshot_state(self, etype: str, normal: set, oversized: set):
        """Cache snapshot state for given `etype`."""
        self._normal_snapshot_eids[etype] |= normal

        self._normal_snapshot_eids[etype] -= oversized
        self._oversized_snapshot_eids[etype] |= oversized

    def save_snapshots(self, etype: str, snapshots: list[dict], ctime: datetime):
        """
        Saves a list of snapshots of current master documents.

        All snapshots must belong to same entity type.

        Will move snapshots to oversized snapshots if the maintained bucket is too large.
        For better understanding, see `save_snapshot()`.
        """
        # Check `etype`
        self._assert_etype_exists(etype)

        for snapshot in snapshots:
            snapshot["_time_created"] = ctime

        snapshot_col = self._snapshots_col_name(etype)
        os_col = self._oversized_snapshots_col_name(etype)

        snapshots_by_eid = defaultdict(list)
        for snapshot in snapshots:
            if "eid" not in snapshot:
                continue
            if "_id" in snapshot:
                del snapshot["_id"]
            snapshots_by_eid[snapshot["eid"]].append(snapshot)

        # Find out if any of the snapshots are oversized
        normal, oversized = self._get_snapshot_state(etype, set(snapshots_by_eid.keys()))

        upserts = []
        update_originals = []
        oversized_inserts = []
        oversized_updates = []

        # A normal snapshot, shift the last snapshot to history and update last
        for eid in normal:
            upserts.append(
                UpdateOne(
                    self._snapshot_bucket_eid_filter(eid)
                    | {"count": {"$lt": self._snapshot_bucket_size}},
                    {
                        "$set": {"last": snapshots_by_eid[eid][-1]},
                        "$push": {"history": {"$each": snapshots_by_eid[eid], "$position": 0}},
                        "$inc": {"count": len(snapshots_by_eid[eid])},
                        "$setOnInsert": {
                            "_id": self._snapshot_bucket_id(eid, ctime),
                            "_time_created": ctime,
                            "oversized": False,
                            "latest": True,
                        },
                    },
                    upsert=True,
                )
            )
            update_originals.append(snapshots_by_eid[eid])

        # Snapshot is already marked as oversized
        for eid in oversized:
            oversized_inserts.extend(snapshots_by_eid[eid])
            oversized_updates.append(
                UpdateOne(
                    self._snapshot_bucket_eid_filter(eid),
                    {"$set": {"last": snapshots_by_eid[eid][-1]}},
                )
            )

        new_oversized = set()

        if upserts:
            try:
                res = self._db[snapshot_col].bulk_write(upserts, ordered=False)

                # Unset latest snapshots if new snapshots were inserted
                if res.upserted_count > 0:
                    unset_latest_updates = []
                    for upsert_id in res.upserted_ids.values():
                        eid = upsert_id.rsplit("_#", maxsplit=1)[0]
                        unset_latest_updates.append(
                            UpdateMany(
                                self._snapshot_bucket_eid_filter(eid)
                                | {"latest": True, "count": self._snapshot_bucket_size},
                                {"$unset": {"latest": 1}},
                            )
                        )
                    up_res = self._db[snapshot_col].bulk_write(unset_latest_updates)
                    if up_res.modified_count != res.upserted_count:
                        self.log.info(
                            "Upserted the first snapshot for %d entities.",
                            res.upserted_count - up_res.modified_count,
                        )

                if res.modified_count + res.upserted_count != len(upserts):
                    self.log.error(
                        "Some snapshots were not updated, %s != %s",
                        res.modified_count + res.upserted_count,
                        len(upserts),
                    )
            except (BulkWriteError, OperationFailure) as e:
                self.log.info("Update of snapshots failed, will retry with oversize.")
                failed_indexes = [
                    err["index"]
                    for err in e.details["writeErrors"]
                    if err["code"] == BSON_OBJECT_TOO_LARGE
                ]
                failed_snapshots = (update_originals[i] for i in failed_indexes)
                for eid_snapshots in failed_snapshots:
                    eid = eid_snapshots[0]["eid"]
                    failed_snapshots = sorted(
                        eid_snapshots, key=lambda s: s["_time_created"], reverse=True
                    )
                    self._migrate_to_oversized_snapshot(etype, eid, failed_snapshots[0])
                    oversized_inserts.extend(failed_snapshots[1:])
                    new_oversized.add(eid)

                if any(err["code"] != BSON_OBJECT_TOO_LARGE for err in e.details["writeErrors"]):
                    # Some other error occurred
                    raise e
            except Exception as e:
                raise DatabaseError(f"Upsert of snapshots failed: {str(e)[:2048]}") from e

        # Update the oversized snapshots
        if oversized_inserts:
            try:
                if oversized_updates:
                    self._db[snapshot_col].bulk_write(oversized_updates)
                self._db[os_col].insert_many(oversized_inserts)
            except Exception as e:
                raise DatabaseError(f"Insert of snapshots failed: {str(e)[:2048]}") from e

        # Cache the new state
        self._cache_snapshot_state(etype, set(), new_oversized)

    def _get_metadata_id(self, module: str, time: datetime, worker_id: Optional[int] = None) -> str:
        """Generates unique metadata id based on `module`, `time` and the worker index."""
        worker_id = self._process_index if worker_id is None else worker_id
        return f"{module}{time.strftime('%Y-%m-%dT%H:%M:%S.%fZ')}w{worker_id}"

    def save_metadata(self, time: datetime, metadata: dict, worker_id: Optional[int] = None):
        """Saves metadata dict under the caller module and passed timestamp."""
        module = get_caller_id()
        metadata["_id"] = self._get_metadata_id(module, time, worker_id)
        metadata["#module"] = module
        metadata["#time_created"] = time
        metadata["#last_update"] = datetime.now()
        try:
            self._db["#metadata"].insert_one(metadata)
            self.log.debug("Inserted metadata %s: %s", metadata["_id"], metadata)
        except Exception as e:
            raise DatabaseError(f"Insert of metadata failed: {e}\n{metadata}") from e

    def update_metadata(
        self, time: datetime, metadata: dict, increase: dict = None, worker_id: Optional[int] = None
    ):
        """Updates existing metadata of caller module and passed timestamp."""
        module = get_caller_id()
        metadata_id = self._get_metadata_id(module, time, worker_id)
        metadata["#last_update"] = datetime.now()

        changes = {"$set": metadata} if increase is None else {"$set": metadata, "$inc": increase}

        try:
            res = self._db["#metadata"].update_one({"_id": metadata_id}, changes, upsert=True)
            self.log.debug(
                "Updated metadata %s, changes: %s, result: %s", metadata_id, changes, res.raw_result
            )
        except Exception as e:
            raise DatabaseError(f"Update of metadata failed: {e}\n{metadata_id}, {changes}") from e

    def get_observation_history(
        self,
        etype: str,
        attr_name: str,
        eid: str,
        t1: datetime = None,
        t2: datetime = None,
        sort: int = None,
    ) -> list[dict]:
        """Get full (or filtered) history of observation attribute.

        This method is useful for displaying `eid`'s history on web.
        Also used to feed data into `get_timeseries_history()`.

        Args:
            etype: entity type
            attr_name: name of attribute
            eid: id of entity, to which data-points correspond
            t1: left value of time interval (inclusive)
            t2: right value of time interval (inclusive)
            sort: sort by timestamps - 0: ascending order by t1, 1: descending order by t2,
                None: don't sort
        Returns:
            list of dicts (reduced datapoints)
        """
        t1 = datetime.fromtimestamp(0) if t1 is None else t1.replace(tzinfo=None)
        t2 = datetime.now() if t2 is None else t2.replace(tzinfo=None)

        # Get attribute history
        mr = self.get_master_record(etype, eid)
        attr_history = mr.get(attr_name, [])

        # Filter
        attr_history_filtered = [row for row in attr_history if row["t1"] <= t2 and row["t2"] >= t1]

        # Sort
        if sort == 1:
            attr_history_filtered.sort(key=lambda row: row["t1"])
        elif sort == 2:
            attr_history_filtered.sort(key=lambda row: row["t2"], reverse=True)

        return attr_history_filtered

    def get_timeseries_history(
        self,
        etype: str,
        attr_name: str,
        eid: str,
        t1: datetime = None,
        t2: datetime = None,
        sort: int = None,
    ) -> list[dict]:
        """Get full (or filtered) history of timeseries attribute.
        Outputs them in format:
        ```
            [
                {
                    "t1": ...,
                    "t2": ...,
                    "v": {
                        "series1": ...,
                        "series2": ...
                    }
                },
                ...
            ]
        ```
        This method is useful for displaying `eid`'s history on web.

        Args:
            etype: entity type
            attr_name: name of attribute
            eid: id of entity, to which data-points correspond
            t1: left value of time interval (inclusive)
            t2: right value of time interval (inclusive)
            sort: sort by timestamps - `0`: ascending order by `t1`, `1`: descending order by `t2`,
                `None`: don't sort
        Returns:
             list of dicts (reduced datapoints) - each represents just one point at time
        """
        t1 = datetime.fromtimestamp(0) if t1 is None else t1.replace(tzinfo=None)
        t2 = datetime.now() if t2 is None else t2.replace(tzinfo=None)

        attr_history = self.get_observation_history(etype, attr_name, eid, t1, t2, sort)
        if not attr_history:
            return []

        attr_history_split = self._split_timeseries_dps(etype, attr_name, attr_history)

        # Filter out rows outside [t1, t2] interval
        attr_history_filtered = [
            row for row in attr_history_split if row["t1"] <= t2 and row["t2"] >= t1
        ]

        return attr_history_filtered

    def _split_timeseries_dps(
        self, etype: str, attr_name: str, attr_history: list[dict]
    ) -> list[dict]:
        """Helper to split "datapoints" (rows) of timeseries to "datapoints"
        containing just one value per series."""
        attrib_conf = self._db_schema_config.attr(etype, attr_name)
        timeseries_type = attrib_conf.timeseries_type

        result = []

        # Should be ["time"] or ["time_first", "time_last"] for irregular
        # timeseries and [] for regular timeseries
        time_series = list(timeseries_types[timeseries_type]["default_series"].keys())

        # User-defined series
        user_series = list(set(attrib_conf.series.keys()) - set(time_series))

        if not user_series:
            return []

        if timeseries_type == "regular":
            time_step = attrib_conf.timeseries_params.time_step

            for row in attr_history:
                # Length of all series arrays/lists in this row
                values_len = len(row["v"][user_series[0]])

                for i in range(values_len):
                    row_t1 = row["t1"] + i * time_step
                    row_t2 = row_t1 + time_step

                    row_new = {"t1": row_t1, "t2": row_t2, "v": {}}

                    for s in user_series:
                        row_new["v"][s] = row["v"][s][i]

                    result.append(row_new)
        else:
            for row in attr_history:
                # Length of all series arrays/lists in this row
                values_len = len(row["v"][user_series[0]])

                for i in range(values_len):
                    row_t1 = row["v"].get(time_series[0])[i] if len(time_series) >= 1 else None
                    row_t2 = row["v"].get(time_series[1])[i] if len(time_series) >= 2 else row_t1

                    row_new = {"t1": row_t1, "t2": row_t2, "v": {}}

                    for s in user_series:
                        row_new["v"][s] = row["v"][s][i]

                    result.append(row_new)

        return result

    def get_distinct_val_count(self, etype: str, attr: str) -> dict[Any, int]:
        """Counts occurences of distinct values of given attribute in snapshots.

        Returns dictionary mapping value -> count.

        Works for all plain and observation data types except `dict` and `json`.
        """
        self._assert_etype_exists(etype)

        snapshot_col = self._snapshots_col_name(etype)

        # Get attribute specification
        try:
            attr_spec = self._db_schema_config.attr(etype, attr)
        except KeyError as e:
            raise DatabaseError(f"Attribute '{attr}' does not exist") from e

        if attr_spec.t not in AttrType.PLAIN | AttrType.OBSERVATIONS:
            raise DatabaseError(f"Attribute '{attr}' isn't plain or observations")

        # Attribute data type must be primitive, array<T> or set<T>
        if attr_spec.data_type.root in ("dict", "json"):
            raise DatabaseError(
                f"Data type '{attr_spec.data_type}' of attribute '{attr}' is not processable"
            )

        # Build aggregation query
        attr_path = "$last." + attr
        unwinding = []

        # Unwind array-like and multi value attributes
        # If attribute is multi value array, unwind twice
        if "array" in attr_spec.data_type.root or "set" in attr_spec.data_type.root:
            unwinding.append({"$unwind": attr_path})
        if attr_spec.t == AttrType.OBSERVATIONS and attr_spec.multi_value:
            unwinding.append({"$unwind": attr_path})

        # Group
        agg_query_group_id = attr_path
        if "link" in attr_spec.data_type.root:
            agg_query_group_id += ".eid"

        agg_query = [
            {"$match": {"latest": True}},
            *unwinding,
            {"$group": {"_id": agg_query_group_id, "count": {"$sum": 1}}},
            {"$sort": {"_id": 1, "count": -1}},
        ]
        # Run aggregation
        distinct_counts_cur = self._db[snapshot_col].aggregate(agg_query)

        distinct_counts = {x["_id"]: x["count"] for x in distinct_counts_cur}

        if None in distinct_counts:
            del distinct_counts[None]

        return distinct_counts

    def move_raw_to_archive(self, etype: str):
        """Rename the current raw collection to archive collection.

        Multiple archive collections can exist for one entity type,
        though they are exported and dropped over time.
        """
        raw_col_name = self._raw_col_name(etype)
        archive_col_name = self._get_new_archive_col_name(etype)
        try:
            if self._db.list_collection_names(filter={"name": raw_col_name}):
                self._db[raw_col_name].rename(archive_col_name)
                return archive_col_name
            return None
        except Exception as e:
            raise DatabaseError(f"Move of raw collection failed: {e}") from e

    def get_archive_summary(self, etype: str, before: datetime) -> Optional[dict]:
        collection_summaries = []
        for archive_col in self._archive_col_names(etype):
            result_cursor = self._get_archive_summary(archive_col, before=before)
            for summary in result_cursor:
                collection_summaries.append(summary)
        if not collection_summaries:
            return None
        min_date = min(x["earliest"] for x in collection_summaries)
        max_date = max(x["latest"] for x in collection_summaries)
        total_dps = sum(x["count"] for x in collection_summaries)
        return {"earliest": min_date, "latest": max_date, "count": total_dps}

    def _get_archive_summary(self, archive_col_name: str, before: datetime) -> CommandCursor:
        try:
            return self._db[archive_col_name].aggregate(
                [
                    {"$match": {"t1": {"$lt": before}}},
                    {
                        "$group": {
                            "_id": "date_range",
                            "earliest": {"$min": "$t1"},
                            "latest": {"$max": "$t1"},
                            "count": {"$sum": 1},
                        }
                    },
                ]
            )
        except Exception as e:
            raise DatabaseError(f"Archive collection {archive_col_name} fetch failed: {e}") from e

    def get_archive(self, etype: str, after: datetime, before: datetime) -> Iterator[Cursor]:
        """Get archived raw datapoints where `t1` is in <`after`, `before`).

        All plain datapoints will be returned (default).
        """
        query_filter = {"$or": [{"t1": {"$gte": after, "$lt": before}}, {"t1": None}]}
        for archive_col in self._archive_col_names(etype):
            try:
                yield self._db[archive_col].find(query_filter)
            except Exception as e:
                raise DatabaseError(f"Archive collection {archive_col} fetch failed: {e}") from e

    def delete_old_archived_dps(self, etype: str, before: datetime) -> Iterator[DeleteResult]:
        """Delete archived raw datapoints older than `before`.

        Deletes all plain datapoints if `plain` is `True` (default).
        """
        query_filter = {"$or": [{"t1": {"$lt": before}}, {"t1": None}]}
        for archive_col in self._archive_col_names(etype):
            try:
                yield self._db[archive_col].delete_many(query_filter)
            except Exception as e:
                raise DatabaseError(f"Delete of old datapoints failed: {e}") from e

    def drop_empty_archives(self, etype: str) -> int:
        """Drop empty archive collections."""
        dropped_count = 0
        for archive_col in self._archive_col_names(etype):
            try:
                col = self._db[archive_col]
                if col.estimated_document_count() < 1000 and col.count_documents({}) == 0:
                    col.drop()
                    dropped_count += 1
            except Exception as e:
                raise DatabaseError(f"Drop of empty archive failed: {e}") from e
        return dropped_count

    def delete_old_snapshots(self, etype: str, t_old: datetime) -> int:
        """Delete old snapshots.

        Periodically called for all `etype`s from HistoryManager.
        """
        snapshot_col_name = self._snapshots_col_name(etype)
        os_snapshot_col_name = self._oversized_snapshots_col_name(etype)
        deleted = 0
        try:
            res = self._db[snapshot_col_name].delete_many(
                {"_time_created": {"$lte": t_old - self._bucket_delta}},
            )
            deleted += res.deleted_count * self._snapshot_bucket_size
            res = self._db[os_snapshot_col_name].delete_many({"_time_created": {"$lt": t_old}})
            deleted += res.deleted_count
        except Exception as e:
            raise DatabaseError(f"Delete of olds snapshots failed: {e}") from e
        return deleted

    def get_module_cache(self, override_called_id: Optional[str] = None):
        """Return a persistent cache collection for given module name.

        Module name is determined automatically, but you can override it.
        """
        module = override_called_id or get_caller_id()
        self.log.debug("Cache collection access: %s", module)
        return self._db[f"#cache#{module}"]

    def get_estimated_entity_count(self, entity_type: str) -> int:
        """Get count of entities of given type."""
        master_col = self._master_col_name(entity_type)
        return self._db[master_col].estimated_document_count()



import logging
import queue
import threading
import time
from datetime import datetime
from functools import partial

from dp3.common.config import PlatformConfig
from dp3.common.task import DataPointTask, parse_data_point_task
from dp3.task_processing.task_executor import TaskExecutor

from ..common.callback_registrar import CallbackRegistrar
from .task_queue import TaskQueueReader, TaskQueueWriter


class TaskDistributor:
    """
    TaskDistributor uses task queues to distribute tasks between all running processes.

    Tasks are assigned to worker processes based on hash of entity key, so each
    entity is always processed by the same worker. Therefore, all requests
    modifying a particular entity are done sequentially and no locking is
    necessary.

    Tasks that are assigned to the current process are passed to `task_executor` for execution.

    Args:
        platform_config: Platform config
        task_executor: Instance of TaskExecutor
        registrar: Interface for callback registration
        daemon_stop_lock: Lock used to control when the program stops. (see [dp3.worker][])
    """

    def __init__(
        self,
        task_executor: TaskExecutor,
        platform_config: PlatformConfig,
        registrar: CallbackRegistrar,
        daemon_stop_lock: threading.Lock,
    ) -> None:
        assert (
            0 <= platform_config.process_index < platform_config.num_processes
        ), "process index must be smaller than number of processes"

        self.log = logging.getLogger("TaskDistributor")

        self.process_index = platform_config.process_index
        self.num_processes = platform_config.num_processes
        self.model_spec = platform_config.model_spec
        self.daemon_stop_lock = daemon_stop_lock

        self.rabbit_params = platform_config.config.get("processing_core.msg_broker", {})

        self.entity_types = list(
            platform_config.config.get("db_entities").keys()
        )  # List of configured entity types

        self.running = False

        # List of worker threads for processing the update requests
        self._worker_threads = []
        self.num_threads = platform_config.config.get("processing_core.worker_threads", 8)

        # Internal queues for each worker
        self._queues = [queue.Queue(10) for _ in range(self.num_threads)]

        # Connections to main task queue
        # Reader - reads tasks from a pair of queues (one pair per process)
        # and distributes them to worker threads
        self._task_queue_reader = TaskQueueReader(
            callback=self._distribute_task,
            parse_task=partial(parse_data_point_task, model_spec=self.model_spec),
            app_name=platform_config.app_name,
            worker_index=self.process_index,
            rabbit_config=self.rabbit_params,
        )
        registrar.scheduler_register(self._task_queue_reader.watchdog, second="20,50")

        # Writer - allows modules to write new tasks
        self._task_queue_writer = TaskQueueWriter(
            platform_config.app_name, self.num_processes, self.rabbit_params
        )
        self.task_executor = task_executor
        # Object to store thread-local data (e.g. worker-thread index)
        # (each thread sees different object contents)
        self._current_thread_data = threading.local()

        # Number of restarts of threads by watchdog
        self._watchdog_restarts = 0
        # Register watchdog to scheduler
        registrar.scheduler_register(self._watchdog, second="*/30")

    def start(self) -> None:
        """Run the worker threads and start consuming from TaskQueue."""
        self.log.info("Connecting to RabbitMQ")
        self._task_queue_reader.connect()
        self._task_queue_reader.check()  # check presence of needed queues
        self._task_queue_writer.connect()
        self._task_queue_writer.check()  # check presence of needed exchanges

        self.log.info(f"Starting {self.num_threads} worker threads")
        self.running = True
        self._worker_threads = [
            threading.Thread(
                target=self._worker_func, args=(i,), name=f"Worker-{self.process_index}-{i}"
            )
            for i in range(self.num_threads)
        ]
        for worker in self._worker_threads:
            worker.start()

        self.log.info("Starting consuming tasks from main queue")
        self._task_queue_reader.start()

    def stop(self) -> None:
        """Stop the worker threads."""
        self.log.info("Waiting for worker threads to finish their current tasks ...")
        # Thread for printing debug messages about worker status
        threading.Thread(target=self._dbg_worker_status_print, daemon=True).start()

        # Stop receiving new tasks from global queue
        self._task_queue_reader.stop()

        # Signalize stop to worker threads
        self.running = False

        # Wait until all workers stopped
        for worker in self._worker_threads:
            worker.join()

        self._task_queue_reader.disconnect()
        self._task_queue_writer.disconnect()

        # Cleanup
        self._worker_threads = []

    def _distribute_task(self, msg_id, task: DataPointTask):
        """
        Puts given task into local queue of the corresponding thread.

        Called by TaskQueueReader when a new task is received from the global queue.
        """
        # Distribute tasks to worker threads by hash of (etype,eid)
        index = hash(task.routing_key()) % self.num_threads
        self._queues[index].put((msg_id, task))

    def _worker_func(self, thread_index):
        """
        Main worker function.

        Run as a separate thread. Read its local task queue and calls
        processing function of TaskExecutor to process each task.

        Tasks are assigned to workers based on hash of entity key, so each
        entity is always processed by the same worker. Therefore, all requests
        modifying a particular entity are done sequentially and no locking is
        necessary.
        """
        # Store index to thread-local variable
        self._current_thread_data.index = thread_index

        my_queue = self._queues[thread_index]

        # Read and process tasks in a loop.
        # Exit immediately after self.running is set to False,
        # it's not a problem if there are any more tasks waiting in the queue
        # - they won't be acknowledged, so they will be re-delivered after restart.
        while self.running:
            # Get message from thread's local queue
            try:
                task_tuple = my_queue.get(block=True, timeout=1)
            except queue.Empty:
                continue  # check self.running again

            msg_id, task = task_tuple

            # Acknowledge receipt of the task (regardless of success/failre of its processing)
            if not self._task_queue_reader.ack(msg_id):
                self.log.warning("Acking message %s failed, will not process.", msg_id)
                continue

            # Process the task
            start_time = datetime.now()
            try:
                created, new_tasks = self.task_executor.process_task(task)
            except Exception:
                self.log.error(f"Error has occurred during processing task: {task}")
                raise

            duration = (datetime.now() - start_time).total_seconds()
            # self.log.debug("Task {} finished in {:.3f} seconds.".format(msg_id, duration))
            if duration > 1.0:
                self.log.debug(
                    "Task {} took {} seconds: {}/{} {}".format(
                        msg_id,
                        duration,
                        task.etype,
                        task.eid,
                        " (new record created)" if created else "",
                    )
                )

            self.push_new_tasks(new_tasks)

    def push_new_tasks(self, new_tasks):
        """Push new tasks (resulting from hooks) to the priority queue.

        (priority queue is not limited in size, so put_task() never blocks; the normal
        queue has limited size, so if it was used here, a deadlock could occur if all
        workers try to push new tasks to a full queue)
        """
        try:
            for task in new_tasks:
                self._task_queue_writer.put_task(task, priority=True)
        except Exception as e:
            self.log.error(f"Failed to push tasks created from hooks: {e}")

    def _watchdog(self):
        """
        Check whether all workers are running and restart them if not.

        Should be called periodically by scheduler.
        Stop whole program after 20 restarts of threads.
        """
        for i, worker in enumerate(self._worker_threads):
            if not worker.is_alive():
                if self._watchdog_restarts < 20:
                    self.log.error(f"Thread {worker.name} is dead, restarting.")
                    worker.join()
                    new_thread = threading.Thread(
                        target=self._worker_func, args=(i,), name=f"Worker-{self.process_index}-{i}"
                    )
                    self._worker_threads[i] = new_thread
                    new_thread.start()
                    self._watchdog_restarts += 1
                else:
                    self.log.critical(
                        f"Thread {worker.name} is dead,"
                        " more than 20 restarts attempted, giving up..."
                    )
                    self.daemon_stop_lock.release()  # Exit program
                    break

    def _dbg_worker_status_print(self):
        """
        Print status of workers and the request queue every 5 seconds.

        Should be run as a separate (daemon) thread.
        Exits when all workers have finished.
        """
        ttl = 10  # Wait for 10 seconds until printing starts
        while True:
            # Check if all workers are dead every second
            time.sleep(1)
            ttl -= 1
            alive_workers = [w for w in self._worker_threads if w.is_alive()]
            if not alive_workers:
                return

            if ttl == 0:
                # Print info and reset counter to 5 seconds
                self.log.info(f"{len(alive_workers)} worker threads alive")
                ttl = 5



import logging
from typing import Callable

from event_count_logger import DummyEventGroup

from dp3.common.config import PlatformConfig
from dp3.common.task import DataPointTask, task_context
from dp3.common.types import EventGroupType
from dp3.database.database import DatabaseError, EntityDatabase
from dp3.task_processing.task_hooks import (
    TaskAttrHooksContainer,
    TaskEntityHooksContainer,
    TaskGenericHooksContainer,
)


class TaskExecutor:
    """
    TaskExecutor manages updates of entity records,
    which are being read from task queue (via parent
    [`TaskDistributor`][dp3.task_processing.task_distributor.TaskDistributor])

    Args:
        db: Instance of EntityDatabase
        platform_config: Current platform configuration.
    """

    def __init__(
        self,
        db: EntityDatabase,
        platform_config: PlatformConfig,
        elog: EventGroupType,
        elog_by_src: EventGroupType,
    ) -> None:
        # initialize task distribution

        self.log = logging.getLogger("TaskExecutor")

        # Get list of configured entity types
        self.entity_types = list(platform_config.model_spec.entities.keys())
        self.log.debug(f"Configured entity types: {self.entity_types}")

        self.model_spec = platform_config.model_spec
        self.db = db

        # Event logging
        self.elog = elog
        self.elog_by_src = elog_by_src
        # Print warning if some event group is not configured
        not_configured_groups = []
        if isinstance(self.elog, DummyEventGroup):
            not_configured_groups.append("te")
        if isinstance(self.elog_by_src, DummyEventGroup):
            not_configured_groups.append("tasks_by_src")
        if not_configured_groups:
            self.log.warning(
                "EventCountLogger: No configuration for event group(s) "
                f"'{','.join(not_configured_groups)}' found, "
                "such events will not be logged (check event_logging.yml)"
            )

        # Hooks
        self._task_generic_hooks = TaskGenericHooksContainer(self.log, self.elog)
        self._task_entity_hooks = {}
        self._task_attr_hooks = {}

        for entity in self.model_spec.entities:
            self._task_entity_hooks[entity] = TaskEntityHooksContainer(
                entity, self.model_spec, self.log, self.elog
            )

        for entity, attr in self.model_spec.attributes:
            attr_type = self.model_spec.attributes[entity, attr].t
            self._task_attr_hooks[entity, attr] = TaskAttrHooksContainer(
                entity, attr, attr_type, self.model_spec, self.log, self.elog
            )

    def register_task_hook(self, hook_type: str, hook: Callable):
        """Registers one of available task hooks

        See: [`TaskGenericHooksContainer`][dp3.task_processing.task_hooks.TaskGenericHooksContainer]
        in `task_hooks.py`
        """
        self._task_generic_hooks.register(hook_type, hook)

    def register_entity_hook(self, hook_type: str, hook: Callable, entity: str):
        """Registers one of available task entity hooks

        See: [`TaskEntityHooksContainer`][dp3.task_processing.task_hooks.TaskEntityHooksContainer]
        in `task_hooks.py`
        """
        self._task_entity_hooks[entity].register(hook_type, hook)

    def register_attr_hook(self, hook_type: str, hook: Callable, entity: str, attr: str):
        """Registers one of available task attribute hooks

        See: [`TaskAttrHooksContainer`][dp3.task_processing.task_hooks.TaskAttrHooksContainer]
        in `task_hooks.py`
        """
        self._task_attr_hooks[entity, attr].register(hook_type, hook)

    def process_task(self, task: DataPointTask) -> tuple[bool, list[DataPointTask]]:
        """
        Main processing function - push datapoint values, running all registered hooks.

        Args:
            task: Task object to process.
        Returns:
            True if a new record was created, False otherwise,
            and a list of new tasks created by hooks
        """
        self.log.debug(f"Received new task {task.etype}/{task.eid}, starting processing!")

        new_tasks = []

        # Run on_task_start hook
        self._task_generic_hooks.run_on_start(task)

        # Check existence of etype
        if task.etype not in self.entity_types:
            self.log.error(f"Task {task.etype}/{task.eid}: Unknown entity type!")
            self.elog.log("task_processing_error")
            return False, new_tasks

        # Check existence of eid
        try:
            ekey_exists = self.db.ekey_exists(task.etype, task.eid)
        except DatabaseError as e:
            self.log.error(f"Task {task.etype}/{task.eid}: DB error: {e}")
            self.elog.log("task_processing_error")
            return False, new_tasks

        new_entity = not ekey_exists

        if task.delete:
            return self._process_delete(task, new_entity)

        if new_entity:
            # Run allow_entity_creation hook
            if not self._task_entity_hooks[task.etype].run_allow_creation(task.eid, task):
                self.log.debug(
                    f"Task {task.etype}/{task.eid}: hooks decided not to create new eid record"
                )
                return False, new_tasks

            # Run on_entity_creation hook
            new_tasks += self._task_entity_hooks[task.etype].run_on_creation(task.eid, task)

        # Extend TTL
        if task.ttl_tokens:
            self.db.extend_ttl(task.etype, task.eid, task.ttl_tokens)

        # Insert into database
        try:
            self.db.insert_datapoints(task.etype, task.eid, task.data_points, new_entity=new_entity)
            self.log.debug(f"Task {task.etype}/{task.eid}: All changes written to DB")
        except DatabaseError as e:
            self.log.error(f"Task {task.etype}/{task.eid}: DB error: {e}")
            self.elog.log("task_processing_error")
            return False, new_tasks

        # Run attribute hooks
        for dp in task.data_points:
            new_tasks += self._task_attr_hooks[dp.etype, dp.attr].run_on_new(dp.eid, dp)

        # Log the processed task
        self.elog.log("task_processed")
        for dp in task.data_points:
            if dp.src:
                self.elog_by_src.log(dp.src)
        if new_entity:
            self.elog.log("record_created")

        self.log.debug(f"Secondary modules created {len(new_tasks)} new tasks.")

        return new_entity, new_tasks

    def refresh_on_entity_creation(
        self, etype: str, worker_id: int, worker_cnt: int
    ) -> list[DataPointTask]:
        """Runs the `on_entity_creation` hook for all entities of given type.

        Returns:
            List of new tasks created by hooks
        """
        if etype not in self.entity_types:
            self.log.error(f"Refresh {etype} on_entity_creation: Unknown entity type!")
            self.elog.log("task_processing_error")

        new_tasks = []

        projection = {"_id": True}
        for master_record in self.db.get_worker_master_records(
            worker_id, worker_cnt, etype, projection=projection
        ):
            with task_context(self.model_spec):
                task = DataPointTask(etype=etype, eid=master_record["_id"])
                self.log.debug(f"Refreshing {etype}/{task.eid}")
                new_tasks += self._task_entity_hooks[task.etype].run_on_creation(task.eid, task)

        return new_tasks

    def _process_delete(
        self, task: DataPointTask, new_entity: bool
    ) -> tuple[bool, list[DataPointTask]]:
        """Helper function to `process_task` for processing delete tasks."""
        if new_entity:
            self.log.debug(
                "Task %s/%s: Master record doesn't exist, nothing to delete.",
                task.etype,
                task.eid,
            )
            return False, []
        try:
            self.log.debug("Task %s/%s: Deleting entity.", task.etype, task.eid)
            self.db.delete_eid(task.etype, task.eid)
        except DatabaseError as e:
            self.log.error(f"Task {task.etype}/{task.eid}: DB error: {e}")
            self.elog.log("task_processing_error")
        return False, []



import logging
import time
from collections import defaultdict
from datetime import datetime
from logging import Logger

import pymongo
from pymongo import DeleteOne, InsertOne
from pymongo.collection import Collection
from pymongo.database import Database

from dp3.common.attrspec import ID_REGEX, AttrSpecType, AttrType
from dp3.common.config import HierarchicalDict, ModelSpec
from dp3.common.utils import batched

# number of seconds to wait for the i-th attempt to reconnect after error
RECONNECT_DELAYS = [1, 2, 5, 10, 30]
# current database schema version
SCHEMA_VERSION = 3

# Collections belonging to entity
# Used when deleting no-longer existing entity.
ENTITY_COLLECTIONS = ["{}#master", "{}#raw", "{}#snapshots"]


class SchemaCleaner:
    """
    Maintains a history of `model_spec` defined schema, updates the database when needed.

    Args:
        db: Target database mapping. (not just a database connection)
        schema_col: Collection where schema history is stored.
        model_spec: Model specification.
        log: Logger instance to serve as parent. If not provided, a new logger will be created.
    """

    def __init__(
        self,
        db: Database,
        schema_col: Collection,
        model_spec: ModelSpec,
        config: HierarchicalDict,
        log: Logger = None,
    ):
        if log is None:
            self.log = logging.getLogger("SchemaCleaner")
        else:
            self.log = log.getChild("SchemaCleaner")

        self._db = db
        self._model_spec = model_spec
        self._config = config
        self.schemas = schema_col

        self.storage = {
            "snapshot_bucket_size": self._config.get("database.storage.snapshot_bucket_size", 32)
        }

        self.migrations = {2: self.migrate_schema_2_to_3}

    def get_current_schema_doc(self, infer: bool = False) -> dict:
        """
        Args:
            infer: Whether to infer the schema if it is not found in the database.
        Returns:
            Current schema
        """
        schema_doc = self.schemas.find_one({}, sort=[("_id", pymongo.DESCENDING)])
        if schema_doc is None and infer:
            self.log.info("No schema found, inferring initial schema from database")
            schema = self.infer_current_schema()
            schema_doc = {
                "_id": 0,
                "schema": schema,
                "storage": self.storage,
                "version": SCHEMA_VERSION,
            }
        return schema_doc

    def safe_update_schema(self):
        """
        Checks whether schema saved in database is up-to-date and updates it if necessary.

        Will NOT perform any changes in master collections on conflicting model changes,
        but will raise an exception instead.
        Any such changes must be performed via the CLI.

        As this method modifies the schema collection, it should be called only by the main worker.

        The schema collection format is as follows:

        ```py
            {
                "_id": int,
                "schema": { ... }, # (1)!
                "storage": {
                    "snapshot_bucket_size": int
                },
                "version": int
            }
        ```

        1. see [schema construction][dp3.database.schema_cleaner.SchemaCleaner.construct_schema_doc]

        Raises:
            ValueError: If conflicting changes are detected.
        """
        db_schema, config_schema, updates, deleted_entites = self.get_schema_status()

        if db_schema["version"] != config_schema["version"]:
            self.log.warning(
                "Schema version mismatch: %s (DB) != %s (config)",
                db_schema["version"],
                config_schema["version"],
            )
            return self._error_on_conflict()

        if db_schema["storage"] != config_schema["storage"]:
            self.log.warning(
                "DB storage settings mismatch: %s (DB) != %s (config)",
                db_schema["storage"],
                config_schema["storage"],
            )
            return self._error_on_conflict()

        if db_schema["schema"] == config_schema["schema"]:
            self.log.info("Schema OK!")
        elif not updates and not deleted_entites:
            self.schemas.insert_one(config_schema)
            self.log.info("Updated schema, OK now!")
        else:
            return self._error_on_conflict()

    def _error_on_conflict(self):
        self.log.warning("Schema update that cannot be performed automatically is required.")
        self.log.warning("Please run `dp3 schema-update` to make the changes.")
        raise ValueError("Schema update failed: Conflicting changes detected.")

    def get_schema_status(self) -> tuple[dict, dict, dict, list]:
        """
        Gets the current schema status.
        `database_schema` is the schema document from the database.
        `configuration_schema` is the schema document constructed from the current configuration.
        `updates` is a dictionary of required updates to each entity.

        Returns:
            Tuple of (`database_schema`, `configuration_schema`, `updates`, `deleted_entites`).
        """
        schema_doc = self.get_current_schema_doc(infer=True)

        current_schema = self.construct_schema_doc()

        if (
            schema_doc["schema"] == current_schema
            and schema_doc["version"] == SCHEMA_VERSION
            and schema_doc["storage"] == self.storage
        ):
            return schema_doc, schema_doc, {}, []

        new_schema = {
            "_id": schema_doc["_id"] + 1,
            "schema": current_schema,
            "storage": self.storage,
            "version": SCHEMA_VERSION,
        }
        updates, deleted_entites = self.detect_changes(schema_doc, current_schema)
        return schema_doc, new_schema, updates, deleted_entites

    def detect_changes(
        self, db_schema_doc: dict, current_schema: dict
    ) -> tuple[dict[str, dict[str, dict[str, str]]], list[str]]:
        """
        Detects changes between configured schema and the one saved in the database.

        Args:
            db_schema_doc: Schema document from the database.
            current_schema: Schema from the configuration.

        Returns:
            Tuple of required updates to each entity and list of deleted entites.
        """
        if db_schema_doc["schema"] == current_schema:
            return {}, []

        if db_schema_doc["version"] != SCHEMA_VERSION:
            self.log.info("Schema version changed, skipping detecting changes.")
            return {}, []

        updates = {}
        deleted_entites = []
        for entity, attributes in db_schema_doc["schema"].items():
            # Unset deleted entities
            if entity not in current_schema:
                self.log.info("Schema breaking change: Entity %s was deleted", entity)
                deleted_entites.append(entity)
                continue

            entity_updates = defaultdict(dict)
            current_attributes = current_schema[entity]

            # Unset deleted attributes
            deleted = set(attributes) - set(current_attributes)
            for attr in deleted:
                self.log.info("Schema breaking change: Attribute %s.%s was deleted", entity, attr)
                entity_updates["$unset"][attr] = ""

            for attr, spec in attributes.items():
                if attr in deleted:
                    continue
                model = current_attributes[attr]

                for key, val in spec.items():
                    ref = model[key]
                    if ref == val:
                        continue

                    self.log.info(
                        "Schema breaking change: '%s' of %s.%s changed: %s -> %s",
                        key,
                        entity,
                        attr,
                        val,
                        ref,
                    )
                    entity_updates["$unset"][attr] = ""

            if entity_updates:
                updates[entity] = entity_updates

        return updates, deleted_entites

    def execute_updates(
        self, updates: dict[str, dict[str, dict[str, str]]], deleted_entites: list[str]
    ):
        # Delete entities
        for entity in deleted_entites:
            try:
                for col_placeholder in ENTITY_COLLECTIONS:
                    col = col_placeholder.format(entity)
                    self.log.info("%s: Deleting collection: %s", entity, col)
                    self._db[col].drop()
            except Exception as e:
                raise ValueError(f"Schema update failed: {e}") from e

        # Update attributes
        for entity, entity_updates in updates.items():
            try:
                self.log.info(
                    "%s: Performing master collection update: %s", entity, dict(entity_updates)
                )
                master_name = f"{entity}#master"
                res = self._db[master_name].update_many({}, entity_updates)
                self.log.debug("%s: Updated %s master records.", entity, res.modified_count)
            except Exception as e:
                raise ValueError(f"Schema update failed: {e}") from e

    def infer_current_schema(self) -> dict:
        """
        Infers schema from current database state.

        Will detect the attribute type, i.e. whether it is
        a plain, timeseries or observations attribute.

        All other properties will be set based on configuration.

        Returns:
            Dictionary with inferred schema.
        """

        cols = self._db.list_collections(filter={"name": {"$regex": r"^[^#]+#master$"}})
        col_names = [col["name"] for col in cols]

        entities = defaultdict(dict)
        for name in col_names:
            entity = name.split("#", maxsplit=1)[0]
            self.log.debug("Inferring schema for '%s':", entity)
            self.log.debug("%s: Testing attributes for history", entity)
            res = self._db[name].aggregate(
                [
                    # Get each attribute as a separate document
                    {"$project": {"items": {"$objectToArray": "$$ROOT"}}},
                    {"$unwind": "$items"},
                    # Check if attribute has history
                    {"$set": {"array": {"$isArray": "$items.v"}}},
                    # Group by attribute name
                    {"$group": {"_id": "$items.k", "array": {"$addToSet": "$array"}}},
                    # Filter out non-attribute properties
                    {"$match": {"_id": {"$regex": ID_REGEX}}},
                    {"$match": {"_id": {"$regex": r"^(?!_id$)"}}},
                ]
            )
            attributes = entities[entity]
            undecided = []
            for doc in res:
                if doc["array"] == [False]:
                    attributes[doc["_id"]] = {"t": AttrType.PLAIN.value}
                elif doc["array"] == [True]:
                    attributes[doc["_id"]] = {
                        "t": AttrType.TIMESERIES.value | AttrType.OBSERVATIONS.value
                    }
                    undecided.append(doc["_id"])
                else:
                    self.log.error("Unknown attribute type: %s.%s", entity, doc["_id"])

            if not undecided:
                continue

            self.log.debug("%s: Testing observations", entity)
            res = self._db[name].aggregate(
                [
                    # Get each attribute as a separate document
                    {"$project": {"items": {"$objectToArray": "$$ROOT"}}},
                    {"$unwind": "$items"},
                    # Check if attribute has confidence (i.e. is observation)
                    {"$match": {"items.k": {"$in": undecided}, "items.v.c": {"$exists": True}}},
                    # Group by attribute name
                    {"$group": {"_id": "$items.k"}},
                ]
            )
            for doc in res:
                attributes[doc["_id"]]["t"] = AttrType.OBSERVATIONS.value
                undecided.remove(doc["_id"])

            if not undecided:
                continue

            self.log.debug("%s: Testing timeseries", entity)
            res = self._db[name].aggregate(
                [
                    # Get each attribute as a separate document
                    {"$project": {"items": {"$objectToArray": "$$ROOT"}}},
                    {"$unwind": "$items"},
                    # Check if attribute has no confidence (i.e. is timeseries)
                    {
                        "$match": {
                            "items.k": {"$in": undecided},
                            "items.v.c": {"$exists": False},
                            "items.v.0": {"$exists": True},
                        }
                    },
                    # Group by attribute name
                    {"$group": {"_id": "$items.k"}},
                ]
            )
            for doc in res:
                attributes[doc["_id"]]["t"] = AttrType.TIMESERIES.value
                undecided.remove(doc["_id"])

            if not undecided:
                continue

            self.log.debug("Resolving undecided attributes: %s", undecided)
            for attr, spec in attributes.items():
                if attr in undecided and (entity, attr) in self._model_spec.attributes:
                    config_spec = self._model_spec.attr(entity, attr)
                    self.log.debug("Assuming configuration type correct for %s.%s", entity, attr)
                    spec["t"] = config_spec.t.value

        for entity, attributes in entities.items():
            for attr, spec in attributes.items():
                if (entity, attr) in self._model_spec.attributes:
                    config_spec = self._model_spec.attr(entity, attr)
                    config_dict = self._construct_spec_dict(config_spec)
                    config_dict.update(spec)
                    spec.update(config_dict)
                self.log.info("Inferred schema for %s.%s: %s", entity, attr, spec)

        return dict(entities)

    def construct_schema_doc(self):
        """
        Constructs schema document from current schema configuration.

        The schema document format is as follows:

        ```py
            {
                "entity_type": {
                    "attribute": {
                        "t": 1 | 2 | 4,
                        "data_type": <data_type.type_info> | None
                        "timeseries_type": <timeseries_type> | None
                        "series": dict[str, <data_type.type_info>] | None
                    }
                }
            }
        ```

        where `t` is [attribute type][dp3.common.attrspec.AttrType],
        and `data_type` is the data type string.
        `timeseries_type` is one of `regular`, `irregular` or `irregular_intervals`.
        `series` is a dictionary of series names and their data types.

        `timeseries_type` and `series` are present only for timeseries attributes,
        `data_type` is present only for plain and observations attributes.
        """
        return {
            entity_type: {
                attr: self._construct_spec_dict(spec) for attr, spec in attributes.items()
            }
            for entity_type, attributes in self._model_spec.entity_attributes.items()
        }

    @staticmethod
    def _construct_spec_dict(spec: AttrSpecType):
        res = {"t": spec.t.value, "data_type": None, "timeseries_type": None, "series": None}
        if spec.t in AttrType.PLAIN | AttrType.OBSERVATIONS:
            res["data_type"] = spec.data_type.type_info
        elif spec.t == AttrType.TIMESERIES:
            res["timeseries_type"] = spec.timeseries_type
            res["series"] = {s: sspec.data_type.type_info for s, sspec in spec.series.items()}
        return res

    def await_updated(self):
        """
        Checks whether schema saved in database is up-to-date and awaits its update
        by the main worker on mismatch.
        """
        self.log.info("Fetching current schema")
        schema_doc = self.get_current_schema_doc()
        for delay in RECONNECT_DELAYS:
            if schema_doc is not None:
                break
            self.log.info("Schema not found, will retry in %s seconds", delay)
            time.sleep(delay)
            schema_doc = self.get_current_schema_doc()

        if schema_doc is None:
            raise ValueError("Unable to establish schema")

        db_schema, config_schema, _updates, _deleted_entites = self.get_schema_status()
        for delay in RECONNECT_DELAYS:
            if db_schema == config_schema:
                break
            self.log.info("Schema mismatch, will await update by main worker in %s seconds", delay)
            time.sleep(delay)
            db_schema, config_schema, _updates, _deleted_entites = self.get_schema_status()

        if db_schema != config_schema:
            raise ValueError("Unable to establish matching schema")

        self.log.info("Schema OK!")

    def get_index_name_by_keys(self, collection: str, keys: dict) -> str:
        """
        Gets index name by keys.

        Args:
            collection: Collection name.
            keys: Index keys.
        Returns:
            Index name.
        """
        for index in self._db[collection].list_indexes():
            if index["key"] == keys:
                return index["name"]
        raise ValueError(f"Index not found for {collection} with keys {keys}")

    def fix_latest_bucket_index(self, collection: str, new_bucket_size: int):
        index = self.get_index_name_by_keys(collection, {"_id": -1, "count": 1})
        self.log.info("Dropping index %s.%s", collection, index)
        self._db[collection].drop_index(index)
        name = self._db[collection].create_index(
            [("_id", pymongo.DESCENDING), ("count", pymongo.ASCENDING)],
            partialFilterExpression={"count": {"$lt": new_bucket_size}},
            background=True,
        )
        self.log.info("Created index %s.%s", collection, name)

    def update_storage(self, prev_storage: dict, curr_storage: dict):
        """
        Updates storage settings in schema.

        Args:
            prev_storage: Previous storage settings.
            curr_storage: Current storage settings.
        """
        self.log.info("Updating storage settings")

        if prev_storage["snapshot_bucket_size"] != curr_storage["snapshot_bucket_size"]:
            self.log.info(
                "Snapshot bucket size changed: %s -> %s",
                prev_storage["snapshot_bucket_size"],
                curr_storage["snapshot_bucket_size"],
            )

            for entity in self._model_spec.entities:
                snapshot_col = f"{entity}#snapshots"

                # Bucket size decreased, fix indexes before updating
                if prev_storage["snapshot_bucket_size"] > curr_storage["snapshot_bucket_size"]:
                    self.fix_latest_bucket_index(snapshot_col, curr_storage["snapshot_bucket_size"])

                self.log.info("Updating %s", entity)
                self._db[snapshot_col].update_many(
                    {"oversized": False},
                    {"$set": {"count": curr_storage["snapshot_bucket_size"]}},
                )

                # Bucket size increased, fix indexes after updating
                if prev_storage["snapshot_bucket_size"] <= curr_storage["snapshot_bucket_size"]:
                    self.fix_latest_bucket_index(snapshot_col, curr_storage["snapshot_bucket_size"])

    def migrate(self, schema_doc: dict) -> dict:
        """
        Migrates schema to the latest version.

        Args:
            schema_doc: Schema to migrate.
        Returns:
            Migrated schema.
        """
        if schema_doc["version"] == SCHEMA_VERSION:
            return schema_doc

        for version in range(schema_doc["version"], SCHEMA_VERSION):
            self.log.info("Migrating schema from version %s to %s", version, version + 1)
            schema_doc = self.migrations[version](schema_doc)
            self.schemas.insert_one(schema_doc)

        return schema_doc

    def migrate_schema_2_to_3(self, schema) -> dict:
        """
        Migrates schema from version 2 to version 3.

        This migration adds the storage setting for snapshot bucket size to the schema.

        Args:
            schema: Schema to migrate.
        Returns:
            Migrated schema.
        """
        # Set created time for all snapshots, append last to history
        bucket_size = self._config.get("database.storage.snapshot_bucket_size", 32)

        for entity in self._model_spec.entities:
            self.log.info("Migrating %s", entity)
            snapshot_col = self._db[f"{entity}#snapshots"]
            for doc in self._db[f"{entity}#snapshots"].find(
                {"_id": {"$not": {"$regex": r"_#\d+$"}}}
            ):
                if doc.get("oversized", False):
                    ctime = doc["last"].get("_time_created", datetime.now())
                    snapshot_col.bulk_write(
                        [
                            InsertOne(
                                {
                                    **doc,
                                    "_id": f"{doc['_id']}_#{int(ctime.timestamp())}",
                                    "_time_created": ctime,
                                    "count": 0,
                                }
                            ),
                            DeleteOne({"_id": doc["_id"]}),
                        ]
                    )
                    continue

                to_insert = []
                for snapshot_bucket in batched([doc["last"]] + doc["history"], bucket_size):
                    ctime = snapshot_bucket[-1].get(
                        "_time_created",
                    )
                    to_insert.append(
                        {
                            "_id": f"{doc['_id']}_#{int(ctime.timestamp())}",
                            "last": snapshot_bucket[0],
                            "history": snapshot_bucket,
                            "_time_created": ctime,
                            "count": len(snapshot_bucket),
                            "oversized": False,
                        }
                    )
                snapshot_col.insert_many(to_insert)
                snapshot_col.delete_one({"_id": doc["_id"]})

        # Update schema
        schema["_id"] += 1
        schema["version"] = 3
        schema["storage"] = {"snapshot_bucket_size": bucket_size}

        return schema
